from enum import StrEnum


class ListingState(StrEnum):
    PRIVATE = "PRIVATE"
    PUBLIC = "PUBLIC"
    READ_ONLY = "READ_ONLY"
    DELISTED = "DELISTED"


class OrderType(StrEnum):
    LIMIT_ORDER = "LIMIT_ORDER"
    CANCEL_ORDER = "CANCEL_ORDER"
    EXCHANGE_INITIATED_CANCELLATION = "EXCHANGE_INITIATED_CANCELLATION"


class OrderState(StrEnum):
    RECEIVED = "RECEIVED"
    PENDING = "PENDING"
    OPEN = "OPEN"
    PARTIAL = "PARTIAL"
    COMPLETED = "COMPLETED"
    CANCELLED = "CANCELLED"
    REJECTED = "REJECTED"


class OrderSide(StrEnum):
    BID = "BID"
    ASK = "ASK"


class TradeState(StrEnum):
    PENDING = "PENDING"
    CLEARED = "CLEARED"
    REJECTED = "REJECTED"



"""MarginAccount contract binding and data structures."""

# This module has been generated using pyabigen v0.2.16

import typing

import eth_typing
import hexbytes
import web3
from web3 import types
from web3.contract import contract

from .types import PositionData


class MarginAccount:
    """MarginAccount contract binding.

    Parameters
    ----------
    w3 : web3.Web3
    address : eth_typing.ChecksumAddress
        The address of a deployed MarginAccount contract.
    """

    _contract: contract.Contract

    def __init__(
        self,
        w3: web3.Web3,
        address: eth_typing.ChecksumAddress,
    ):
        self._contract = w3.eth.contract(
            address=address,
            abi=ABI,
        )

    @property
    def Deposit(self) -> contract.ContractEvent:
        """Binding for `event Deposit` on the MarginAccount contract."""
        return self._contract.events.Deposit

    @property
    def Initialized(self) -> contract.ContractEvent:
        """Binding for `event Initialized` on the MarginAccount contract."""
        return self._contract.events.Initialized

    @property
    def IntentAuthorized(self) -> contract.ContractEvent:
        """Binding for `event IntentAuthorized` on the MarginAccount contract."""
        return self._contract.events.IntentAuthorized

    @property
    def IntentRevoked(self) -> contract.ContractEvent:
        """Binding for `event IntentRevoked` on the MarginAccount contract."""
        return self._contract.events.IntentRevoked

    @property
    def OwnershipTransferred(self) -> contract.ContractEvent:
        """Binding for `event OwnershipTransferred` on the MarginAccount contract."""
        return self._contract.events.OwnershipTransferred

    @property
    def PositionUpdated(self) -> contract.ContractEvent:
        """Binding for `event PositionUpdated` on the MarginAccount contract."""
        return self._contract.events.PositionUpdated

    @property
    def Withdraw(self) -> contract.ContractEvent:
        """Binding for `event Withdraw` on the MarginAccount contract."""
        return self._contract.events.Withdraw

    def add_allowed(
        self,
        addr: eth_typing.ChecksumAddress,
    ) -> contract.ContractFunction:
        """Binding for `addAllowed` on the MarginAccount contract.

        Parameters
        ----------
        addr : eth_typing.ChecksumAddress

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.addAllowed(
            addr,
        )

    def admin(
        self,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> eth_typing.ChecksumAddress:
        """Binding for `admin` on the MarginAccount contract.

        Parameters
        ----------
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        eth_typing.ChecksumAddress
        """
        return_value = self._contract.functions.admin().call(
            block_identifier=block_identifier
        )
        return eth_typing.ChecksumAddress(return_value)

    def allowed(
        self,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> typing.List[eth_typing.ChecksumAddress]:
        """Binding for `allowed` on the MarginAccount contract.

        Parameters
        ----------
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        typing.List[eth_typing.ChecksumAddress]
        """
        return_value = self._contract.functions.allowed().call(
            block_identifier=block_identifier
        )
        return [
            eth_typing.ChecksumAddress(return_value_elem)
            for return_value_elem in return_value
        ]

    def authorize(
        self,
        intent_account: eth_typing.ChecksumAddress,
    ) -> contract.ContractFunction:
        """Binding for `authorize` on the MarginAccount contract.

        Parameters
        ----------
        intent_account : eth_typing.ChecksumAddress

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.authorize(
            intent_account,
        )

    def authorized(
        self,
        margin_account_id: eth_typing.ChecksumAddress,
        intent_account: eth_typing.ChecksumAddress,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> bool:
        """Binding for `authorized` on the MarginAccount contract.

        Parameters
        ----------
        margin_account_id : eth_typing.ChecksumAddress
        intent_account : eth_typing.ChecksumAddress
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        bool
        """
        return_value = self._contract.functions.authorized(
            margin_account_id,
            intent_account,
        ).call(block_identifier=block_identifier)
        return bool(return_value)

    def capital(
        self,
        margin_account_id: eth_typing.ChecksumAddress,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> int:
        """Binding for `capital` on the MarginAccount contract.

        Parameters
        ----------
        margin_account_id : eth_typing.ChecksumAddress
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        int
        """
        return_value = self._contract.functions.capital(
            margin_account_id,
        ).call(block_identifier=block_identifier)
        return int(return_value)

    def clearing(
        self,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> eth_typing.ChecksumAddress:
        """Binding for `clearing` on the MarginAccount contract.

        Parameters
        ----------
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        eth_typing.ChecksumAddress
        """
        return_value = self._contract.functions.clearing().call(
            block_identifier=block_identifier
        )
        return eth_typing.ChecksumAddress(return_value)

    def collateral_asset(
        self,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> eth_typing.ChecksumAddress:
        """Binding for `collateralAsset` on the MarginAccount contract.

        Parameters
        ----------
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        eth_typing.ChecksumAddress
        """
        return_value = self._contract.functions.collateralAsset().call(
            block_identifier=block_identifier
        )
        return eth_typing.ChecksumAddress(return_value)

    def deposit(
        self,
        amount: int,
    ) -> contract.ContractFunction:
        """Binding for `deposit` on the MarginAccount contract.

        Parameters
        ----------
        amount : int

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.deposit(
            amount,
        )

    def deposit_for(
        self,
        margin_account_id: eth_typing.ChecksumAddress,
        amount: int,
    ) -> contract.ContractFunction:
        """Binding for `depositFor` on the MarginAccount contract.

        Parameters
        ----------
        margin_account_id : eth_typing.ChecksumAddress
        amount : int

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.depositFor(
            margin_account_id,
            amount,
        )

    def initialize(
        self,
        clearing_: eth_typing.ChecksumAddress,
    ) -> contract.ContractFunction:
        """Binding for `initialize` on the MarginAccount contract.

        Parameters
        ----------
        clearing_ : eth_typing.ChecksumAddress

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.initialize(
            clearing_,
        )

    def mae(
        self,
        margin_account_id: eth_typing.ChecksumAddress,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> int:
        """Binding for `mae` on the MarginAccount contract.

        Parameters
        ----------
        margin_account_id : eth_typing.ChecksumAddress
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        int
        """
        return_value = self._contract.functions.mae(
            margin_account_id,
        ).call(block_identifier=block_identifier)
        return int(return_value)

    def mma(
        self,
        margin_account_id: eth_typing.ChecksumAddress,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> int:
        """Binding for `mma` on the MarginAccount contract.

        Parameters
        ----------
        margin_account_id : eth_typing.ChecksumAddress
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        int
        """
        return_value = self._contract.functions.mma(
            margin_account_id,
        ).call(block_identifier=block_identifier)
        return int(return_value)

    def mmu(
        self,
        margin_account_id: eth_typing.ChecksumAddress,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> int:
        """Binding for `mmu` on the MarginAccount contract.

        Parameters
        ----------
        margin_account_id : eth_typing.ChecksumAddress
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        int
        """
        return_value = self._contract.functions.mmu(
            margin_account_id,
        ).call(block_identifier=block_identifier)
        return int(return_value)

    def owner(
        self,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> eth_typing.ChecksumAddress:
        """Binding for `owner` on the MarginAccount contract.

        Parameters
        ----------
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        eth_typing.ChecksumAddress
        """
        return_value = self._contract.functions.owner().call(
            block_identifier=block_identifier
        )
        return eth_typing.ChecksumAddress(return_value)

    def pnl(
        self,
        margin_account_id: eth_typing.ChecksumAddress,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> int:
        """Binding for `pnl` on the MarginAccount contract.

        Parameters
        ----------
        margin_account_id : eth_typing.ChecksumAddress
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        int
        """
        return_value = self._contract.functions.pnl(
            margin_account_id,
        ).call(block_identifier=block_identifier)
        return int(return_value)

    def position_count(
        self,
        margin_account_id: eth_typing.ChecksumAddress,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> int:
        """Binding for `positionCount` on the MarginAccount contract.

        Parameters
        ----------
        margin_account_id : eth_typing.ChecksumAddress
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        int
        """
        return_value = self._contract.functions.positionCount(
            margin_account_id,
        ).call(block_identifier=block_identifier)
        return int(return_value)

    def position_data(
        self,
        margin_account_id: eth_typing.ChecksumAddress,
        product_id: hexbytes.HexBytes,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> PositionData:
        """Binding for `positionData` on the MarginAccount contract.

        Parameters
        ----------
        margin_account_id : eth_typing.ChecksumAddress
        product_id : hexbytes.HexBytes
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        PositionData
        """
        return_value = self._contract.functions.positionData(
            margin_account_id,
            product_id,
        ).call(block_identifier=block_identifier)
        return PositionData(
            hexbytes.HexBytes(return_value[0]),
            int(return_value[1]),
            int(return_value[2]),
            int(return_value[3]),
            int(return_value[4]),
        )

    def position_pn_l(
        self,
        margin_account: eth_typing.ChecksumAddress,
        product_id: hexbytes.HexBytes,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> int:
        """Binding for `positionPnL` on the MarginAccount contract.

        Parameters
        ----------
        margin_account : eth_typing.ChecksumAddress
        product_id : hexbytes.HexBytes
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        int
        """
        return_value = self._contract.functions.positionPnL(
            margin_account,
            product_id,
        ).call(block_identifier=block_identifier)
        return int(return_value)

    def position_quantity(
        self,
        margin_account_id: eth_typing.ChecksumAddress,
        product_id: hexbytes.HexBytes,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> int:
        """Binding for `positionQuantity` on the MarginAccount contract.

        Parameters
        ----------
        margin_account_id : eth_typing.ChecksumAddress
        product_id : hexbytes.HexBytes
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        int
        """
        return_value = self._contract.functions.positionQuantity(
            margin_account_id,
            product_id,
        ).call(block_identifier=block_identifier)
        return int(return_value)

    def positions(
        self,
        margin_account_id: eth_typing.ChecksumAddress,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> typing.List[hexbytes.HexBytes]:
        """Binding for `positions` on the MarginAccount contract.

        Parameters
        ----------
        margin_account_id : eth_typing.ChecksumAddress
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        typing.List[hexbytes.HexBytes]
        """
        return_value = self._contract.functions.positions(
            margin_account_id,
        ).call(block_identifier=block_identifier)
        return [
            hexbytes.HexBytes(return_value_elem) for return_value_elem in return_value
        ]

    def remove_allowed(
        self,
        addr: eth_typing.ChecksumAddress,
    ) -> contract.ContractFunction:
        """Binding for `removeAllowed` on the MarginAccount contract.

        Parameters
        ----------
        addr : eth_typing.ChecksumAddress

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.removeAllowed(
            addr,
        )

    def renounce_ownership(
        self,
    ) -> contract.ContractFunction:
        """Binding for `renounceOwnership` on the MarginAccount contract.

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.renounceOwnership()

    def revoke_authorization(
        self,
        intent_account: eth_typing.ChecksumAddress,
    ) -> contract.ContractFunction:
        """Binding for `revokeAuthorization` on the MarginAccount contract.

        Parameters
        ----------
        intent_account : eth_typing.ChecksumAddress

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.revokeAuthorization(
            intent_account,
        )

    def transfer_ownership(
        self,
        new_owner: eth_typing.ChecksumAddress,
    ) -> contract.ContractFunction:
        """Binding for `transferOwnership` on the MarginAccount contract.

        Parameters
        ----------
        new_owner : eth_typing.ChecksumAddress

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.transferOwnership(
            new_owner,
        )

    def withdraw(
        self,
        amount: int,
    ) -> contract.ContractFunction:
        """Binding for `withdraw` on the MarginAccount contract.

        Parameters
        ----------
        amount : int

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.withdraw(
            amount,
        )

    def withdrawable(
        self,
        margin_account_id: eth_typing.ChecksumAddress,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> int:
        """Binding for `withdrawable` on the MarginAccount contract.

        Parameters
        ----------
        margin_account_id : eth_typing.ChecksumAddress
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        int
        """
        return_value = self._contract.functions.withdrawable(
            margin_account_id,
        ).call(block_identifier=block_identifier)
        return int(return_value)


ABI = typing.cast(
    eth_typing.ABI,
    [
        {
            "type": "function",
            "name": "addAllowed",
            "inputs": [{"name": "addr", "type": "address", "internalType": "address"}],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "admin",
            "inputs": [],
            "outputs": [{"name": "", "type": "address", "internalType": "address"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "allowed",
            "inputs": [],
            "outputs": [{"name": "", "type": "address[]", "internalType": "address[]"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "authorize",
            "inputs": [
                {"name": "intentAccount", "type": "address", "internalType": "address"}
            ],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "authorized",
            "inputs": [
                {
                    "name": "marginAccountId",
                    "type": "address",
                    "internalType": "address",
                },
                {"name": "intentAccount", "type": "address", "internalType": "address"},
            ],
            "outputs": [{"name": "", "type": "bool", "internalType": "bool"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "capital",
            "inputs": [
                {
                    "name": "marginAccountId",
                    "type": "address",
                    "internalType": "address",
                }
            ],
            "outputs": [{"name": "", "type": "int256", "internalType": "int256"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "clearing",
            "inputs": [],
            "outputs": [
                {
                    "name": "",
                    "type": "address",
                    "internalType": "contract IMarginAccountFacet",
                }
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "collateralAsset",
            "inputs": [],
            "outputs": [{"name": "", "type": "address", "internalType": "address"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "deposit",
            "inputs": [
                {"name": "amount", "type": "uint256", "internalType": "uint256"}
            ],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "depositFor",
            "inputs": [
                {
                    "name": "marginAccountId",
                    "type": "address",
                    "internalType": "address",
                },
                {"name": "amount", "type": "uint256", "internalType": "uint256"},
            ],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "initialize",
            "inputs": [
                {"name": "clearing_", "type": "address", "internalType": "address"}
            ],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "mae",
            "inputs": [
                {
                    "name": "marginAccountId",
                    "type": "address",
                    "internalType": "address",
                }
            ],
            "outputs": [{"name": "", "type": "int256", "internalType": "int256"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "mma",
            "inputs": [
                {
                    "name": "marginAccountId",
                    "type": "address",
                    "internalType": "address",
                }
            ],
            "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "mmu",
            "inputs": [
                {
                    "name": "marginAccountId",
                    "type": "address",
                    "internalType": "address",
                }
            ],
            "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "owner",
            "inputs": [],
            "outputs": [{"name": "", "type": "address", "internalType": "address"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "pnl",
            "inputs": [
                {
                    "name": "marginAccountId",
                    "type": "address",
                    "internalType": "address",
                }
            ],
            "outputs": [{"name": "", "type": "int256", "internalType": "int256"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "positionCount",
            "inputs": [
                {
                    "name": "marginAccountId",
                    "type": "address",
                    "internalType": "address",
                }
            ],
            "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "positionData",
            "inputs": [
                {
                    "name": "marginAccountId",
                    "type": "address",
                    "internalType": "address",
                },
                {"name": "productId", "type": "bytes32", "internalType": "bytes32"},
            ],
            "outputs": [
                {
                    "name": "",
                    "type": "tuple",
                    "internalType": "struct PositionData",
                    "components": [
                        {
                            "name": "productId",
                            "type": "bytes32",
                            "internalType": "bytes32",
                        },
                        {
                            "name": "quantity",
                            "type": "int256",
                            "internalType": "int256",
                        },
                        {
                            "name": "costBasis",
                            "type": "int256",
                            "internalType": "int256",
                        },
                        {
                            "name": "maintenanceMargin",
                            "type": "uint256",
                            "internalType": "uint256",
                        },
                        {"name": "pnl", "type": "int256", "internalType": "int256"},
                    ],
                }
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "positionPnL",
            "inputs": [
                {"name": "marginAccount", "type": "address", "internalType": "address"},
                {"name": "productId", "type": "bytes32", "internalType": "bytes32"},
            ],
            "outputs": [{"name": "", "type": "int256", "internalType": "int256"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "positionQuantity",
            "inputs": [
                {
                    "name": "marginAccountId",
                    "type": "address",
                    "internalType": "address",
                },
                {"name": "productId", "type": "bytes32", "internalType": "bytes32"},
            ],
            "outputs": [{"name": "", "type": "int256", "internalType": "int256"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "positions",
            "inputs": [
                {
                    "name": "marginAccountId",
                    "type": "address",
                    "internalType": "address",
                }
            ],
            "outputs": [{"name": "", "type": "bytes32[]", "internalType": "bytes32[]"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "removeAllowed",
            "inputs": [{"name": "addr", "type": "address", "internalType": "address"}],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "renounceOwnership",
            "inputs": [],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "revokeAuthorization",
            "inputs": [
                {"name": "intentAccount", "type": "address", "internalType": "address"}
            ],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "transferOwnership",
            "inputs": [
                {"name": "newOwner", "type": "address", "internalType": "address"}
            ],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "withdraw",
            "inputs": [
                {"name": "amount", "type": "uint256", "internalType": "uint256"}
            ],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "withdrawable",
            "inputs": [
                {
                    "name": "marginAccountId",
                    "type": "address",
                    "internalType": "address",
                }
            ],
            "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}],
            "stateMutability": "view",
        },
        {
            "type": "event",
            "name": "Deposit",
            "inputs": [
                {
                    "name": "user",
                    "type": "address",
                    "indexed": True,
                    "internalType": "address",
                },
                {
                    "name": "amount",
                    "type": "uint256",
                    "indexed": False,
                    "internalType": "uint256",
                },
            ],
            "anonymous": False,
        },
        {
            "type": "event",
            "name": "Initialized",
            "inputs": [
                {
                    "name": "version",
                    "type": "uint64",
                    "indexed": False,
                    "internalType": "uint64",
                }
            ],
            "anonymous": False,
        },
        {
            "type": "event",
            "name": "IntentAuthorized",
            "inputs": [
                {
                    "name": "marginAccountId",
                    "type": "address",
                    "indexed": True,
                    "internalType": "address",
                },
                {
                    "name": "intentAccount",
                    "type": "address",
                    "indexed": True,
                    "internalType": "address",
                },
            ],
            "anonymous": False,
        },
        {
            "type": "event",
            "name": "IntentRevoked",
            "inputs": [
                {
                    "name": "marginAccountId",
                    "type": "address",
                    "indexed": True,
                    "internalType": "address",
                },
                {
                    "name": "intentAccount",
                    "type": "address",
                    "indexed": True,
                    "internalType": "address",
                },
            ],
            "anonymous": False,
        },
        {
            "type": "event",
            "name": "OwnershipTransferred",
            "inputs": [
                {
                    "name": "previousOwner",
                    "type": "address",
                    "indexed": True,
                    "internalType": "address",
                },
                {
                    "name": "newOwner",
                    "type": "address",
                    "indexed": True,
                    "internalType": "address",
                },
            ],
            "anonymous": False,
        },
        {
            "type": "event",
            "name": "PositionUpdated",
            "inputs": [
                {
                    "name": "marginAccountId",
                    "type": "address",
                    "indexed": True,
                    "internalType": "address",
                },
                {
                    "name": "productId",
                    "type": "bytes32",
                    "indexed": True,
                    "internalType": "bytes32",
                },
                {
                    "name": "totalQuantity",
                    "type": "int256",
                    "indexed": False,
                    "internalType": "int256",
                },
                {
                    "name": "costBasis",
                    "type": "int256",
                    "indexed": False,
                    "internalType": "int256",
                },
                {
                    "name": "price",
                    "type": "uint256",
                    "indexed": False,
                    "internalType": "uint256",
                },
                {
                    "name": "quantity",
                    "type": "int256",
                    "indexed": False,
                    "internalType": "int256",
                },
            ],
            "anonymous": False,
        },
        {
            "type": "event",
            "name": "Withdraw",
            "inputs": [
                {
                    "name": "user",
                    "type": "address",
                    "indexed": True,
                    "internalType": "address",
                },
                {
                    "name": "amount",
                    "type": "uint256",
                    "indexed": False,
                    "internalType": "uint256",
                },
            ],
            "anonymous": False,
        },
        {"type": "error", "name": "InvalidInitialization", "inputs": []},
        {"type": "error", "name": "NotInitializing", "inputs": []},
        {
            "type": "error",
            "name": "OwnableInvalidOwner",
            "inputs": [{"name": "owner", "type": "address", "internalType": "address"}],
        },
        {
            "type": "error",
            "name": "OwnableUnauthorizedAccount",
            "inputs": [
                {"name": "account", "type": "address", "internalType": "address"}
            ],
        },
        {
            "type": "error",
            "name": "SafeCastOverflowedIntToUint",
            "inputs": [{"name": "value", "type": "int256", "internalType": "int256"}],
        },
        {
            "type": "error",
            "name": "SafeERC20FailedOperation",
            "inputs": [{"name": "token", "type": "address", "internalType": "address"}],
        },
        {
            "type": "error",
            "name": "Unauthorized",
            "inputs": [
                {"name": "account", "type": "address", "internalType": "address"}
            ],
        },
    ],
)



"""Shared test fixtures and helpers for integration tests."""

from datetime import datetime, timezone
from decimal import Decimal
from typing import Any
from unittest.mock import Mock

import ipld_car  # type: ignore (untyped library)
from requests import Response

from afp.dtos import (
    ComponentLink,
    ExchangeParameters,
    ExchangeProductListingSubmission,
    ExchangeProductUpdateSubmission,
    ExtendedMetadata,
    ExtendedMetadataDAG,
    LoginSubmission,
    OrderSubmission,
)
from afp.enums import ListingState, OrderSide, OrderState, OrderType, TradeState
from afp.schemas import (
    ExchangeProduct,
    IntentData,
    MarketDepthData,
    MarketDepthItem,
    OHLCVItem,
    OracleFallback,
    Order,
    OrderFill,
    OutcomePointTimeSeries,
    OutcomeSpaceTimeSeries,
    Trade,
)


# Sample data factories for DTOs


def make_exchange_product(**overrides: Any) -> ExchangeProduct:
    """Create a sample ExchangeProduct with optional field overrides."""
    defaults = {
        "id": "0x1234567890123456789012345678901234567890123456789012345678901234",
        "symbol": "BTCUSD",
        "tick_size": 1,
        "collateral_asset": "0x0000000000000000000000000000000000000001",
        "listing_state": ListingState.PUBLIC,
        "min_price": Decimal("0"),
        "max_price": Decimal("100000"),
    }
    return ExchangeProduct(**{**defaults, **overrides})


def make_intent_data(**overrides: Any) -> IntentData:
    """Create a sample IntentData with optional field overrides."""
    defaults = {
        "trading_protocol_id": "0xabcd",
        "product_id": "0x1234567890123456789012345678901234567890123456789012345678901234",
        "limit_price": Decimal("50000"),
        "quantity": 100,
        "max_trading_fee_rate": Decimal("0.001"),
        "side": OrderSide.BID,
        "good_until_time": 1700000000,
        "nonce": 1,
        "referral": "0x0000000000000000000000000000000000000000",
    }
    return IntentData(**{**defaults, **overrides})


def make_order(**overrides: Any) -> Order:
    """Create a sample Order with optional field overrides."""
    intent_data = make_intent_data()
    defaults = {
        "id": "order-123",
        "type": OrderType.LIMIT_ORDER,
        "timestamp": 1700000000,
        "state": OrderState.OPEN,
        "fill_quantity": 0,
        "intent": {
            "hash": "0x9876543210987654321098765432109876543210987654321098765432109876",
            "margin_account_id": "0x1111111111111111111111111111111111111111",
            "intent_account_id": "0x2222222222222222222222222222222222222222",
            "signature": "0xabcdef",
            "data": intent_data.model_dump(),
        },
    }
    return Order(**{**defaults, **overrides})


def make_trade(**overrides: Any) -> Trade:
    """Create a sample Trade with optional field overrides."""
    defaults = {
        "id": "123",
        "product_id": "0x1234567890123456789012345678901234567890123456789012345678901234",
        "price": Decimal("50000"),
        "timestamp": 1700000000,
        "state": TradeState.CLEARED,
        "transaction_id": "0xabc123",
        "rejection_reason": None,
    }
    return Trade(**{**defaults, **overrides})


def make_order_fill(**overrides: Any) -> OrderFill:
    """Create a sample OrderFill with optional field overrides."""
    defaults = {
        "order": make_order().model_dump(),
        "trade": make_trade().model_dump(),
        "quantity": 50,
        "price": Decimal("50000"),
        "trading_fee_rate": Decimal("0.001"),
    }
    return OrderFill(**{**defaults, **overrides})


def make_market_depth_item(**overrides: Any) -> MarketDepthItem:
    """Create a sample MarketDepthItem with optional field overrides."""
    defaults = {
        "price": Decimal("50000"),
        "quantity": 100,
    }
    return MarketDepthItem(**{**defaults, **overrides})


def make_market_depth_data(**overrides: Any) -> MarketDepthData:
    """Create a sample MarketDepthData with optional field overrides."""
    defaults = {
        "product_id": "0x1234567890123456789012345678901234567890123456789012345678901234",
        "bids": [
            make_market_depth_item(price=Decimal("49900"), quantity=50).model_dump(),
            make_market_depth_item(price=Decimal("49800"), quantity=100).model_dump(),
        ],
        "asks": [
            make_market_depth_item(price=Decimal("50100"), quantity=50).model_dump(),
            make_market_depth_item(price=Decimal("50200"), quantity=100).model_dump(),
        ],
    }
    return MarketDepthData(**{**defaults, **overrides})


def make_ohlcv_item(**overrides: Any) -> OHLCVItem:
    """Create a sample OHLCVItem with optional field overrides."""
    defaults = {
        "timestamp": 1700000000,
        "open": Decimal("50000"),
        "high": Decimal("51000"),
        "low": Decimal("49000"),
        "close": Decimal("50500"),
        "volume": 1000,
    }
    return OHLCVItem(**{**defaults, **overrides})


def make_login_submission(**overrides: Any) -> LoginSubmission:
    """Create a sample LoginSubmission with optional field overrides."""
    defaults = {
        "message": "Login message",
        "signature": "0xabcdef1234567890",
    }
    return LoginSubmission(**{**defaults, **overrides})


def make_exchange_parameters(**overrides: Any) -> ExchangeParameters:
    """Create a sample ExchangeParameters with optional field overrides."""
    defaults = {
        "trading_protocol_id": "0xabcd",
        "maker_trading_fee_rate": Decimal("0.001"),
        "taker_trading_fee_rate": Decimal("0.002"),
    }
    return ExchangeParameters(**{**defaults, **overrides})


def make_order_submission(**overrides: Any) -> OrderSubmission:
    """Create a sample OrderSubmission with optional field overrides."""
    intent_data = make_intent_data()
    defaults = {
        "type": OrderType.LIMIT_ORDER,
        "intent": {
            "hash": "0x9876543210987654321098765432109876543210987654321098765432109876",
            "margin_account_id": "0x1111111111111111111111111111111111111111",
            "intent_account_id": "0x2222222222222222222222222222222222222222",
            "signature": "0xabcdef",
            "data": intent_data.model_dump(),
        },
        "cancellation_data": None,
    }
    return OrderSubmission(**{**defaults, **overrides})


def make_product_listing_submission(
    **overrides: Any,
) -> ExchangeProductListingSubmission:
    """Create a sample ExchangeProductListingSubmission."""
    defaults = {
        "id": "0x1234567890123456789012345678901234567890123456789012345678901234",
    }
    return ExchangeProductListingSubmission(**{**defaults, **overrides})


def make_product_update_submission(**overrides: Any) -> ExchangeProductUpdateSubmission:
    """Create a sample ExchangeProductUpdateSubmission."""
    defaults = {
        "listing_state": ListingState.PUBLIC,
    }
    return ExchangeProductUpdateSubmission(**{**defaults, **overrides})


# IPFS-specific fixtures


def make_outcome_space_time_series(**overrides: Any) -> OutcomeSpaceTimeSeries:
    """Create a sample OutcomeSpaceTimeSeries with optional field overrides."""
    defaults = {
        "fsp_type": "scalar",
        "description": "Test outcome space",
        "base_case": {
            "condition": "value is valid",
            "fsp_resolution": "value",
        },
        "edge_cases": [],
        "units": "USD",
        "source_name": "Test Source",
        "source_uri": "https://example.com/data",
        "frequency": "daily",
        "history_api_spec": None,
    }
    return OutcomeSpaceTimeSeries(**{**defaults, **overrides})


def make_outcome_point_time_series(**overrides: Any) -> OutcomePointTimeSeries:
    """Create a sample OutcomePointTimeSeries with optional field overrides."""
    defaults = {
        "fsp_type": "scalar",
        "observation": {
            "reference_date": "2024-01-01",
            "release_date": "2024-01-02",
        },
    }
    return OutcomePointTimeSeries(**{**defaults, **overrides})


def make_oracle_fallback(**overrides: Any) -> OracleFallback:
    """Create a sample OracleFallback with optional field overrides."""
    defaults = {
        "fallback_time": datetime(2025, 1, 15, tzinfo=timezone.utc),
        "fallback_fsp": Decimal("100"),
    }
    return OracleFallback(**{**defaults, **overrides})


def make_extended_metadata(**overrides: Any) -> ExtendedMetadata:
    """Create a sample ExtendedMetadata with optional field overrides."""
    from afp.schemas import OracleConfig

    defaults = {
        "outcome_space": make_outcome_space_time_series(),
        "outcome_point": make_outcome_point_time_series(),
        "oracle_config": OracleConfig(
            description="Test oracle",
            project_url="https://example.com",
        ),
        "oracle_fallback": make_oracle_fallback(),
    }
    return ExtendedMetadata(**{**defaults, **overrides})


def make_extended_metadata_dag(**overrides: Any) -> ExtendedMetadataDAG:
    """Create a sample ExtendedMetadataDAG with optional field overrides."""
    # Use real valid CIDs (minimum valid base32 CIDs)
    defaults = {
        "outcome_space": ComponentLink(
            data="bafyreiaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
            schema_="bafyreibbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb",
        ).model_dump(),
        "outcome_point": ComponentLink(
            data="bafyreiccccccccccccccccccccccccccccccccccccccccccccccccccccc",
            schema_="bafyreiddddddddddddddddddddddddddddddddddddddddddddddddddddd",
        ).model_dump(),
        "oracle_config": ComponentLink(
            data="bafyreieeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeee",
            schema_="bafyreifffffffffffffffffffffffffffffffffffffffffffffffffffffff",
        ).model_dump(),
        "oracle_fallback": ComponentLink(
            data="bafyreiggggggggggggggggggggggggggggggggggggggggggggggggggggg",
            schema_="bafyreihhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhhh",
        ).model_dump(),
    }
    return ExtendedMetadataDAG(**{**defaults, **overrides})


# Response generators


def make_ndjson_response(items: list[Any]) -> Response:
    """Create a mock Response with iter_lines() returning NDJSON."""
    mock_response = Response()
    mock_response.status_code = 200
    mock_response._content = b""  # type: ignore
    mock_response.iter_lines = Mock(  # type: ignore
        return_value=iter([item.model_dump_json().encode() for item in items])
    )
    return mock_response


def make_error_response(status_code: int, detail: str | list[Any]) -> Response:
    """Generate standardized error responses for various HTTP codes."""
    import requests

    mock_response = Response()
    mock_response.status_code = status_code
    mock_response._content = b""  # type: ignore

    if isinstance(detail, str):
        mock_response.json = Mock(return_value={"detail": detail})  # type: ignore
    else:
        mock_response.json = Mock(return_value={"detail": detail})  # type: ignore

    # Simulate raise_for_status behavior
    http_error = requests.exceptions.HTTPError(response=mock_response)
    mock_response.raise_for_status = Mock(side_effect=http_error)  # type: ignore

    return mock_response


def make_success_response(data: dict[str, Any]) -> Response:
    """Create a mock successful Response."""
    mock_response = Response()
    mock_response.status_code = 200
    mock_response._content = b""  # type: ignore
    mock_response.json = Mock(return_value=data)  # type: ignore
    mock_response.history = []  # type: ignore
    return mock_response


# IPFS-specific helpers


def make_ipfs_car_response(root_cid: str) -> Response:
    """Create a mock IPFS CAR upload response."""
    mock_response = Response()
    mock_response.status_code = 200
    mock_response._content = b""  # type: ignore
    mock_response.json = Mock(  # type: ignore
        return_value={
            "Root": {
                "Cid": {"/": root_cid},
            },
        }
    )
    return mock_response


def make_ipfs_block_response(content: bytes) -> Response:
    """Create a mock IPFS block download response."""
    mock_response = Response()
    mock_response.status_code = 200
    mock_response._content = content  # type: ignore
    return mock_response


def make_car_bytes(blocks: list[ipld_car.Block]) -> bytes:
    """Generate CAR file bytes from blocks."""
    root_cid = blocks[0][0]
    return ipld_car.encode([root_cid], blocks).tobytes()



"""AdminFacet contract binding and data structures."""

# This module has been generated using pyabigen v0.2.16

import typing

import eth_typing
import web3
from web3 import types
from web3.contract import contract


from .types import (
    ClearingConfig,
    MarkPriceConfig,
    FinalSettlementConfig,
    ProductConfig,
    Config,
)


class AdminFacet:
    """AdminFacet contract binding.

    Parameters
    ----------
    w3 : web3.Web3
    address : eth_typing.ChecksumAddress
        The address of a deployed AdminFacet contract.
    """

    _contract: contract.Contract

    def __init__(
        self,
        w3: web3.Web3,
        address: eth_typing.ChecksumAddress,
    ):
        self._contract = w3.eth.contract(
            address=address,
            abi=ABI,
        )

    @property
    def AddressUpdated(self) -> contract.ContractEvent:
        """Binding for `event AddressUpdated` on the AdminFacet contract."""
        return self._contract.events.AddressUpdated

    @property
    def ConfigUpdated(self) -> contract.ContractEvent:
        """Binding for `event ConfigUpdated` on the AdminFacet contract."""
        return self._contract.events.ConfigUpdated

    @property
    def OwnershipTransferred(self) -> contract.ContractEvent:
        """Binding for `event OwnershipTransferred` on the AdminFacet contract."""
        return self._contract.events.OwnershipTransferred

    def clearing_fee_rate(
        self,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> int:
        """Binding for `clearingFeeRate` on the AdminFacet contract.

        Parameters
        ----------
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        int
        """
        return_value = self._contract.functions.clearingFeeRate().call(
            block_identifier=block_identifier
        )
        return int(return_value)

    def closeout_fee_rate(
        self,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> int:
        """Binding for `closeoutFeeRate` on the AdminFacet contract.

        Parameters
        ----------
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        int
        """
        return_value = self._contract.functions.closeoutFeeRate().call(
            block_identifier=block_identifier
        )
        return int(return_value)

    def closeout_reward_rate(
        self,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> int:
        """Binding for `closeoutRewardRate` on the AdminFacet contract.

        Parameters
        ----------
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        int
        """
        return_value = self._contract.functions.closeoutRewardRate().call(
            block_identifier=block_identifier
        )
        return int(return_value)

    def config(
        self,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> Config:
        """Binding for `config` on the AdminFacet contract.

        Parameters
        ----------
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        Config
        """
        return_value = self._contract.functions.config().call(
            block_identifier=block_identifier
        )
        return Config(
            ClearingConfig(int(return_value[0][0]), int(return_value[0][1])),
            MarkPriceConfig(int(return_value[1][0])),
            FinalSettlementConfig(int(return_value[2][0]), int(return_value[2][1])),
            ProductConfig(
                int(return_value[3][0]),
                int(return_value[3][1]),
                int(return_value[3][2]),
            ),
        )

    def get_margin_account_registry(
        self,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> eth_typing.ChecksumAddress:
        """Binding for `getMarginAccountRegistry` on the AdminFacet contract.

        Parameters
        ----------
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        eth_typing.ChecksumAddress
        """
        return_value = self._contract.functions.getMarginAccountRegistry().call(
            block_identifier=block_identifier
        )
        return eth_typing.ChecksumAddress(return_value)

    def get_mark_price_interval(
        self,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> int:
        """Binding for `getMarkPriceInterval` on the AdminFacet contract.

        Parameters
        ----------
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        int
        """
        return_value = self._contract.functions.getMarkPriceInterval().call(
            block_identifier=block_identifier
        )
        return int(return_value)

    def get_product_registry(
        self,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> eth_typing.ChecksumAddress:
        """Binding for `getProductRegistry` on the AdminFacet contract.

        Parameters
        ----------
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        eth_typing.ChecksumAddress
        """
        return_value = self._contract.functions.getProductRegistry().call(
            block_identifier=block_identifier
        )
        return eth_typing.ChecksumAddress(return_value)

    def get_treasury(
        self,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> eth_typing.ChecksumAddress:
        """Binding for `getTreasury` on the AdminFacet contract.

        Parameters
        ----------
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        eth_typing.ChecksumAddress
        """
        return_value = self._contract.functions.getTreasury().call(
            block_identifier=block_identifier
        )
        return eth_typing.ChecksumAddress(return_value)

    def initialize(
        self,
        treasury: eth_typing.ChecksumAddress,
        product_registry: eth_typing.ChecksumAddress,
        margin_account_registry: eth_typing.ChecksumAddress,
    ) -> contract.ContractFunction:
        """Binding for `initialize` on the AdminFacet contract.

        Parameters
        ----------
        treasury : eth_typing.ChecksumAddress
        product_registry : eth_typing.ChecksumAddress
        margin_account_registry : eth_typing.ChecksumAddress

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.initialize(
            treasury,
            product_registry,
            margin_account_registry,
        )

    def max_trading_fee_rate(
        self,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> int:
        """Binding for `maxTradingFeeRate` on the AdminFacet contract.

        Parameters
        ----------
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        int
        """
        return_value = self._contract.functions.maxTradingFeeRate().call(
            block_identifier=block_identifier
        )
        return int(return_value)

    def owner(
        self,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> eth_typing.ChecksumAddress:
        """Binding for `owner` on the AdminFacet contract.

        Parameters
        ----------
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        eth_typing.ChecksumAddress
        """
        return_value = self._contract.functions.owner().call(
            block_identifier=block_identifier
        )
        return eth_typing.ChecksumAddress(return_value)

    def set_config(
        self,
        config_: Config,
    ) -> contract.ContractFunction:
        """Binding for `setConfig` on the AdminFacet contract.

        Parameters
        ----------
        config_ : Config

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.setConfig(
            (
                (
                    config_.clearing_config.clearing_fee_rate,
                    config_.clearing_config.max_trading_fee_rate,
                ),
                (config_.mark_price_config.mark_price_interval),
                (
                    config_.final_settlement_config.closeout_fee_rate,
                    config_.final_settlement_config.closeout_reward_rate,
                ),
                (
                    config_.product_config.clearing_payout_ratio,
                    config_.product_config.maintenance_deposit_interval,
                    config_.product_config.min_product_maintenance_fee,
                ),
            ),
        )

    def set_margin_account_registry(
        self,
        margin_account_registry: eth_typing.ChecksumAddress,
    ) -> contract.ContractFunction:
        """Binding for `setMarginAccountRegistry` on the AdminFacet contract.

        Parameters
        ----------
        margin_account_registry : eth_typing.ChecksumAddress

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.setMarginAccountRegistry(
            margin_account_registry,
        )

    def set_treasury(
        self,
        treasury: eth_typing.ChecksumAddress,
    ) -> contract.ContractFunction:
        """Binding for `setTreasury` on the AdminFacet contract.

        Parameters
        ----------
        treasury : eth_typing.ChecksumAddress

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.setTreasury(
            treasury,
        )

    def transfer_ownership(
        self,
        new_owner: eth_typing.ChecksumAddress,
    ) -> contract.ContractFunction:
        """Binding for `transferOwnership` on the AdminFacet contract.

        Parameters
        ----------
        new_owner : eth_typing.ChecksumAddress

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.transferOwnership(
            new_owner,
        )


ABI = typing.cast(
    eth_typing.ABI,
    [
        {
            "type": "function",
            "name": "clearingFeeRate",
            "inputs": [],
            "outputs": [{"name": "", "type": "uint32", "internalType": "uint32"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "closeoutFeeRate",
            "inputs": [],
            "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "closeoutRewardRate",
            "inputs": [],
            "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "config",
            "inputs": [],
            "outputs": [
                {
                    "name": "",
                    "type": "tuple",
                    "internalType": "struct Config",
                    "components": [
                        {
                            "name": "clearingConfig",
                            "type": "tuple",
                            "internalType": "struct ClearingConfig",
                            "components": [
                                {
                                    "name": "clearingFeeRate",
                                    "type": "uint32",
                                    "internalType": "uint32",
                                },
                                {
                                    "name": "maxTradingFeeRate",
                                    "type": "uint32",
                                    "internalType": "uint32",
                                },
                            ],
                        },
                        {
                            "name": "markPriceConfig",
                            "type": "tuple",
                            "internalType": "struct MarkPriceConfig",
                            "components": [
                                {
                                    "name": "markPriceInterval",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                }
                            ],
                        },
                        {
                            "name": "finalSettlementConfig",
                            "type": "tuple",
                            "internalType": "struct FinalSettlementConfig",
                            "components": [
                                {
                                    "name": "closeoutFeeRate",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                                {
                                    "name": "closeoutRewardRate",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                            ],
                        },
                        {
                            "name": "productConfig",
                            "type": "tuple",
                            "internalType": "struct ProductConfig",
                            "components": [
                                {
                                    "name": "clearingPayoutRatio",
                                    "type": "uint32",
                                    "internalType": "uint32",
                                },
                                {
                                    "name": "maintenanceDepositInterval",
                                    "type": "uint32",
                                    "internalType": "uint32",
                                },
                                {
                                    "name": "minProductMaintenanceFee",
                                    "type": "uint64",
                                    "internalType": "uint64",
                                },
                            ],
                        },
                    ],
                }
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "getMarginAccountRegistry",
            "inputs": [],
            "outputs": [
                {
                    "name": "",
                    "type": "address",
                    "internalType": "contract IMarginAccountRegistry",
                }
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "getMarkPriceInterval",
            "inputs": [],
            "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "getProductRegistry",
            "inputs": [],
            "outputs": [
                {
                    "name": "",
                    "type": "address",
                    "internalType": "contract IProductRegistry",
                }
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "getTreasury",
            "inputs": [],
            "outputs": [{"name": "", "type": "address", "internalType": "address"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "initialize",
            "inputs": [
                {"name": "treasury", "type": "address", "internalType": "address"},
                {
                    "name": "productRegistry",
                    "type": "address",
                    "internalType": "address",
                },
                {
                    "name": "marginAccountRegistry",
                    "type": "address",
                    "internalType": "address",
                },
            ],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "maxTradingFeeRate",
            "inputs": [],
            "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "owner",
            "inputs": [],
            "outputs": [{"name": "", "type": "address", "internalType": "address"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "setConfig",
            "inputs": [
                {
                    "name": "config_",
                    "type": "tuple",
                    "internalType": "struct Config",
                    "components": [
                        {
                            "name": "clearingConfig",
                            "type": "tuple",
                            "internalType": "struct ClearingConfig",
                            "components": [
                                {
                                    "name": "clearingFeeRate",
                                    "type": "uint32",
                                    "internalType": "uint32",
                                },
                                {
                                    "name": "maxTradingFeeRate",
                                    "type": "uint32",
                                    "internalType": "uint32",
                                },
                            ],
                        },
                        {
                            "name": "markPriceConfig",
                            "type": "tuple",
                            "internalType": "struct MarkPriceConfig",
                            "components": [
                                {
                                    "name": "markPriceInterval",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                }
                            ],
                        },
                        {
                            "name": "finalSettlementConfig",
                            "type": "tuple",
                            "internalType": "struct FinalSettlementConfig",
                            "components": [
                                {
                                    "name": "closeoutFeeRate",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                                {
                                    "name": "closeoutRewardRate",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                            ],
                        },
                        {
                            "name": "productConfig",
                            "type": "tuple",
                            "internalType": "struct ProductConfig",
                            "components": [
                                {
                                    "name": "clearingPayoutRatio",
                                    "type": "uint32",
                                    "internalType": "uint32",
                                },
                                {
                                    "name": "maintenanceDepositInterval",
                                    "type": "uint32",
                                    "internalType": "uint32",
                                },
                                {
                                    "name": "minProductMaintenanceFee",
                                    "type": "uint64",
                                    "internalType": "uint64",
                                },
                            ],
                        },
                    ],
                }
            ],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "setMarginAccountRegistry",
            "inputs": [
                {
                    "name": "marginAccountRegistry",
                    "type": "address",
                    "internalType": "address",
                }
            ],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "setTreasury",
            "inputs": [
                {"name": "treasury", "type": "address", "internalType": "address"}
            ],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "transferOwnership",
            "inputs": [
                {"name": "newOwner", "type": "address", "internalType": "address"}
            ],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "event",
            "name": "AddressUpdated",
            "inputs": [
                {
                    "name": "name",
                    "type": "string",
                    "indexed": False,
                    "internalType": "string",
                },
                {
                    "name": "oldAddress",
                    "type": "address",
                    "indexed": False,
                    "internalType": "address",
                },
                {
                    "name": "newAddress",
                    "type": "address",
                    "indexed": False,
                    "internalType": "address",
                },
            ],
            "anonymous": False,
        },
        {
            "type": "event",
            "name": "ConfigUpdated",
            "inputs": [
                {
                    "name": "oldConfig",
                    "type": "tuple",
                    "indexed": False,
                    "internalType": "struct Config",
                    "components": [
                        {
                            "name": "clearingConfig",
                            "type": "tuple",
                            "internalType": "struct ClearingConfig",
                            "components": [
                                {
                                    "name": "clearingFeeRate",
                                    "type": "uint32",
                                    "internalType": "uint32",
                                },
                                {
                                    "name": "maxTradingFeeRate",
                                    "type": "uint32",
                                    "internalType": "uint32",
                                },
                            ],
                        },
                        {
                            "name": "markPriceConfig",
                            "type": "tuple",
                            "internalType": "struct MarkPriceConfig",
                            "components": [
                                {
                                    "name": "markPriceInterval",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                }
                            ],
                        },
                        {
                            "name": "finalSettlementConfig",
                            "type": "tuple",
                            "internalType": "struct FinalSettlementConfig",
                            "components": [
                                {
                                    "name": "closeoutFeeRate",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                                {
                                    "name": "closeoutRewardRate",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                            ],
                        },
                        {
                            "name": "productConfig",
                            "type": "tuple",
                            "internalType": "struct ProductConfig",
                            "components": [
                                {
                                    "name": "clearingPayoutRatio",
                                    "type": "uint32",
                                    "internalType": "uint32",
                                },
                                {
                                    "name": "maintenanceDepositInterval",
                                    "type": "uint32",
                                    "internalType": "uint32",
                                },
                                {
                                    "name": "minProductMaintenanceFee",
                                    "type": "uint64",
                                    "internalType": "uint64",
                                },
                            ],
                        },
                    ],
                },
                {
                    "name": "newConfig",
                    "type": "tuple",
                    "indexed": False,
                    "internalType": "struct Config",
                    "components": [
                        {
                            "name": "clearingConfig",
                            "type": "tuple",
                            "internalType": "struct ClearingConfig",
                            "components": [
                                {
                                    "name": "clearingFeeRate",
                                    "type": "uint32",
                                    "internalType": "uint32",
                                },
                                {
                                    "name": "maxTradingFeeRate",
                                    "type": "uint32",
                                    "internalType": "uint32",
                                },
                            ],
                        },
                        {
                            "name": "markPriceConfig",
                            "type": "tuple",
                            "internalType": "struct MarkPriceConfig",
                            "components": [
                                {
                                    "name": "markPriceInterval",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                }
                            ],
                        },
                        {
                            "name": "finalSettlementConfig",
                            "type": "tuple",
                            "internalType": "struct FinalSettlementConfig",
                            "components": [
                                {
                                    "name": "closeoutFeeRate",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                                {
                                    "name": "closeoutRewardRate",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                            ],
                        },
                        {
                            "name": "productConfig",
                            "type": "tuple",
                            "internalType": "struct ProductConfig",
                            "components": [
                                {
                                    "name": "clearingPayoutRatio",
                                    "type": "uint32",
                                    "internalType": "uint32",
                                },
                                {
                                    "name": "maintenanceDepositInterval",
                                    "type": "uint32",
                                    "internalType": "uint32",
                                },
                                {
                                    "name": "minProductMaintenanceFee",
                                    "type": "uint64",
                                    "internalType": "uint64",
                                },
                            ],
                        },
                    ],
                },
            ],
            "anonymous": False,
        },
        {
            "type": "event",
            "name": "OwnershipTransferred",
            "inputs": [
                {
                    "name": "previousOwner",
                    "type": "address",
                    "indexed": True,
                    "internalType": "address",
                },
                {
                    "name": "newOwner",
                    "type": "address",
                    "indexed": True,
                    "internalType": "address",
                },
            ],
            "anonymous": False,
        },
        {
            "type": "error",
            "name": "InvalidParameter",
            "inputs": [
                {"name": "paramName", "type": "string", "internalType": "string"}
            ],
        },
    ],
)



from os import path

import afp

from hexbytes import HexBytes


def test_PrivateKeyAuthenticator():
    authenticator = afp.PrivateKeyAuthenticator(
        "0x93f2348b2b890be53f55344b7b118b53778440eea0dc84a6f86fb6e7f6bd77a6"
    )

    assert authenticator.address == "0xFfbf2643CF22760AfD3b878BA8aE849c48944Aa5"
    assert authenticator.sign_message(b"foobar") == HexBytes(
        "0x32b31738559341bdcdd56ef3bb38dbbb9f218d3ac7b84c0118e364c1e036dc1c07d82f0bc9b2c6428966b7a2a10689ae047b8e765df3b66f5c16e90e20632b681b"
    )


def test_KeyfileAuthenticator():
    keyfile = path.join(path.dirname(__file__), "assets", "test.key")
    authenticator = afp.KeyfileAuthenticator(keyfile, "foobar")

    assert authenticator.address == "0xFfbf2643CF22760AfD3b878BA8aE849c48944Aa5"
    assert authenticator.sign_message(b"foobar") == HexBytes(
        "0x32b31738559341bdcdd56ef3bb38dbbb9f218d3ac7b84c0118e364c1e036dc1c07d82f0bc9b2c6428966b7a2a10689ae047b8e765df3b66f5c16e90e20632b681b"
    )



import os
from typing import cast
from unittest.mock import Mock

from eth_account.datastructures import SignedTransaction
from eth_typing.evm import ChecksumAddress
from hexbytes import HexBytes
from web3.types import TxParams

os.environ = {}

from afp import auth


NULL_ADDRESS = cast(ChecksumAddress, "0x0000000000000000000000000000000000000000")


class AuthenticatorStub(auth.Authenticator):
    address = NULL_ADDRESS

    def sign_message(self, message: bytes) -> HexBytes:
        return HexBytes("0x")

    def sign_transaction(self, params: TxParams) -> SignedTransaction:
        return cast(SignedTransaction, Mock(raw_transaction=HexBytes("0x")))



"""SystemViewer contract binding and data structures."""

# This module has been generated using pyabigen v0.2.16

import enum
import typing
from dataclasses import dataclass

import eth_typing
import hexbytes
import web3
from web3 import types
from web3.contract import contract


class ProductState(enum.IntEnum):
    """Port of `enum ProductState` on the SystemViewer contract."""

    NOT_EXIST = 0
    PENDING = 1
    LIVE = 2
    TRADEOUT = 3
    FINAL_SETTLEMENT = 4
    EXPIRED = 5


@dataclass
class Settlement:
    """Port of `struct Settlement` on the SystemViewer contract."""

    product_id: hexbytes.HexBytes
    quantity: int
    price: int


@dataclass
class PositionData:
    """Port of `struct PositionData` on the SystemViewer contract."""

    product_id: hexbytes.HexBytes
    quantity: int
    cost_basis: int
    maintenance_margin: int
    pnl: int


@dataclass
class ProductMetadata:
    """Port of `struct ProductMetadata` on the SystemViewer contract."""

    builder: eth_typing.ChecksumAddress
    symbol: str
    description: str


@dataclass
class OracleSpecification:
    """Port of `struct OracleSpecification` on the SystemViewer contract."""

    oracle_address: eth_typing.ChecksumAddress
    fsv_decimals: int
    fsp_alpha: int
    fsp_beta: int
    fsv_calldata: hexbytes.HexBytes


@dataclass
class BaseProduct:
    """Port of `struct BaseProduct` on the SystemViewer contract."""

    metadata: ProductMetadata
    oracle_spec: OracleSpecification
    collateral_asset: eth_typing.ChecksumAddress
    start_time: int
    point_value: int
    price_decimals: int
    extended_metadata: str


@dataclass
class ExpirySpecification:
    """Port of `struct ExpirySpecification` on the SystemViewer contract."""

    earliest_fsp_submission_time: int
    tradeout_interval: int


@dataclass
class PredictionProductV1:
    """Port of `struct PredictionProductV1` on the SystemViewer contract."""

    base: BaseProduct
    expiry_spec: ExpirySpecification
    max_price: int
    min_price: int


@dataclass
class UserMarginAccountData:
    """Port of `struct UserMarginAccountData` on the ISystemViewer contract."""

    collateral_asset: eth_typing.ChecksumAddress
    margin_account_id: eth_typing.ChecksumAddress
    mae: int
    mmu: int
    positions: typing.List[PositionData]


class SystemViewer:
    """SystemViewer contract binding.

    Parameters
    ----------
    w3 : web3.Web3
    address : eth_typing.ChecksumAddress
        The address of a deployed SystemViewer contract.
    """

    _contract: contract.Contract

    def __init__(
        self,
        w3: web3.Web3,
        address: eth_typing.ChecksumAddress,
    ):
        self._contract = w3.eth.contract(
            address=address,
            abi=ABI,
        )

    @property
    def Initialized(self) -> contract.ContractEvent:
        """Binding for `event Initialized` on the SystemViewer contract."""
        return self._contract.events.Initialized

    @property
    def OwnershipTransferred(self) -> contract.ContractEvent:
        """Binding for `event OwnershipTransferred` on the SystemViewer contract."""
        return self._contract.events.OwnershipTransferred

    @property
    def Upgraded(self) -> contract.ContractEvent:
        """Binding for `event Upgraded` on the SystemViewer contract."""
        return self._contract.events.Upgraded

    def upgrade_interface_version(
        self,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> str:
        """Binding for `UPGRADE_INTERFACE_VERSION` on the SystemViewer contract.

        Parameters
        ----------
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        str
        """
        return_value = self._contract.functions.UPGRADE_INTERFACE_VERSION().call(
            block_identifier=block_identifier
        )
        return str(return_value)

    def clearing(
        self,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> eth_typing.ChecksumAddress:
        """Binding for `clearing` on the SystemViewer contract.

        Parameters
        ----------
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        eth_typing.ChecksumAddress
        """
        return_value = self._contract.functions.clearing().call(
            block_identifier=block_identifier
        )
        return eth_typing.ChecksumAddress(return_value)

    def initialize(
        self,
        _clearing: eth_typing.ChecksumAddress,
        _product_registry: eth_typing.ChecksumAddress,
        _margin_account_registry: eth_typing.ChecksumAddress,
    ) -> contract.ContractFunction:
        """Binding for `initialize` on the SystemViewer contract.

        Parameters
        ----------
        _clearing : eth_typing.ChecksumAddress
        _product_registry : eth_typing.ChecksumAddress
        _margin_account_registry : eth_typing.ChecksumAddress

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.initialize(
            _clearing,
            _product_registry,
            _margin_account_registry,
        )

    def mae_by_collateral_asset(
        self,
        collateral_asset: eth_typing.ChecksumAddress,
        accounts: typing.List[eth_typing.ChecksumAddress],
        block_identifier: types.BlockIdentifier = "latest",
    ) -> typing.List[int]:
        """Binding for `maeByCollateralAsset` on the SystemViewer contract.

        Parameters
        ----------
        collateral_asset : eth_typing.ChecksumAddress
        accounts : typing.List[eth_typing.ChecksumAddress]
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        typing.List[int]
        """
        return_value = self._contract.functions.maeByCollateralAsset(
            collateral_asset,
            accounts,
        ).call(block_identifier=block_identifier)
        return [int(return_value_elem) for return_value_elem in return_value]

    def mae_by_collateral_assets(
        self,
        collateral_assets: typing.List[eth_typing.ChecksumAddress],
        accounts: typing.List[eth_typing.ChecksumAddress],
        block_identifier: types.BlockIdentifier = "latest",
    ) -> typing.List[int]:
        """Binding for `maeByCollateralAssets` on the SystemViewer contract.

        Parameters
        ----------
        collateral_assets : typing.List[eth_typing.ChecksumAddress]
        accounts : typing.List[eth_typing.ChecksumAddress]
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        typing.List[int]
        """
        return_value = self._contract.functions.maeByCollateralAssets(
            collateral_assets,
            accounts,
        ).call(block_identifier=block_identifier)
        return [int(return_value_elem) for return_value_elem in return_value]

    def mae_checks(
        self,
        accounts: typing.List[eth_typing.ChecksumAddress],
        settlements: typing.List[Settlement],
        block_identifier: types.BlockIdentifier = "latest",
    ) -> typing.Tuple[typing.List[bool], typing.List[int], typing.List[int]]:
        """Binding for `maeChecks` on the SystemViewer contract.

        Parameters
        ----------
        accounts : typing.List[eth_typing.ChecksumAddress]
        settlements : typing.List[Settlement]
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        typing.List[bool]
        typing.List[int]
        typing.List[int]
        """
        return_value = self._contract.functions.maeChecks(
            accounts,
            [(item.product_id, item.quantity, item.price) for item in settlements],
        ).call(block_identifier=block_identifier)
        return (
            [bool(return_value0_elem) for return_value0_elem in return_value[0]],
            [int(return_value1_elem) for return_value1_elem in return_value[1]],
            [int(return_value2_elem) for return_value2_elem in return_value[2]],
        )

    def margin_account_registry(
        self,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> eth_typing.ChecksumAddress:
        """Binding for `marginAccountRegistry` on the SystemViewer contract.

        Parameters
        ----------
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        eth_typing.ChecksumAddress
        """
        return_value = self._contract.functions.marginAccountRegistry().call(
            block_identifier=block_identifier
        )
        return eth_typing.ChecksumAddress(return_value)

    def owner(
        self,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> eth_typing.ChecksumAddress:
        """Binding for `owner` on the SystemViewer contract.

        Parameters
        ----------
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        eth_typing.ChecksumAddress
        """
        return_value = self._contract.functions.owner().call(
            block_identifier=block_identifier
        )
        return eth_typing.ChecksumAddress(return_value)

    def position_quantities_by_product_id(
        self,
        product_id: hexbytes.HexBytes,
        accounts: typing.List[eth_typing.ChecksumAddress],
        block_identifier: types.BlockIdentifier = "latest",
    ) -> typing.List[int]:
        """Binding for `positionQuantitiesByProductId` on the SystemViewer contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes
        accounts : typing.List[eth_typing.ChecksumAddress]
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        typing.List[int]
        """
        return_value = self._contract.functions.positionQuantitiesByProductId(
            product_id,
            accounts,
        ).call(block_identifier=block_identifier)
        return [int(return_value_elem) for return_value_elem in return_value]

    def position_quantities_by_product_ids(
        self,
        product_ids: typing.List[hexbytes.HexBytes],
        accounts: typing.List[eth_typing.ChecksumAddress],
        block_identifier: types.BlockIdentifier = "latest",
    ) -> typing.List[int]:
        """Binding for `positionQuantitiesByProductIds` on the SystemViewer contract.

        Parameters
        ----------
        product_ids : typing.List[hexbytes.HexBytes]
        accounts : typing.List[eth_typing.ChecksumAddress]
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        typing.List[int]
        """
        return_value = self._contract.functions.positionQuantitiesByProductIds(
            product_ids,
            accounts,
        ).call(block_identifier=block_identifier)
        return [int(return_value_elem) for return_value_elem in return_value]

    def positions_by_collateral_asset(
        self,
        collateral_asset: eth_typing.ChecksumAddress,
        accounts: typing.List[eth_typing.ChecksumAddress],
        block_identifier: types.BlockIdentifier = "latest",
    ) -> typing.List[typing.List[PositionData]]:
        """Binding for `positionsByCollateralAsset` on the SystemViewer contract.

        Parameters
        ----------
        collateral_asset : eth_typing.ChecksumAddress
        accounts : typing.List[eth_typing.ChecksumAddress]
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        typing.List[typing.List[PositionData]]
        """
        return_value = self._contract.functions.positionsByCollateralAsset(
            collateral_asset,
            accounts,
        ).call(block_identifier=block_identifier)
        return [
            [
                PositionData(
                    hexbytes.HexBytes(return_value_elem[0]),
                    int(return_value_elem[1]),
                    int(return_value_elem[2]),
                    int(return_value_elem[3]),
                    int(return_value_elem[4]),
                )
                for return_value_elem in return_value_list0
            ]
            for return_value_list0 in return_value
        ]

    def positions_by_collateral_assets(
        self,
        collateral_assets: typing.List[eth_typing.ChecksumAddress],
        accounts: typing.List[eth_typing.ChecksumAddress],
        block_identifier: types.BlockIdentifier = "latest",
    ) -> typing.List[typing.List[PositionData]]:
        """Binding for `positionsByCollateralAssets` on the SystemViewer contract.

        Parameters
        ----------
        collateral_assets : typing.List[eth_typing.ChecksumAddress]
        accounts : typing.List[eth_typing.ChecksumAddress]
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        typing.List[typing.List[PositionData]]
        """
        return_value = self._contract.functions.positionsByCollateralAssets(
            collateral_assets,
            accounts,
        ).call(block_identifier=block_identifier)
        return [
            [
                PositionData(
                    hexbytes.HexBytes(return_value_elem[0]),
                    int(return_value_elem[1]),
                    int(return_value_elem[2]),
                    int(return_value_elem[3]),
                    int(return_value_elem[4]),
                )
                for return_value_elem in return_value_list0
            ]
            for return_value_list0 in return_value
        ]

    def product_details(
        self,
        product_ids: typing.List[hexbytes.HexBytes],
        block_identifier: types.BlockIdentifier = "latest",
    ) -> typing.List[PredictionProductV1]:
        """Binding for `productDetails` on the SystemViewer contract.

        Parameters
        ----------
        product_ids : typing.List[hexbytes.HexBytes]
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        typing.List[PredictionProductV1]
        """
        return_value = self._contract.functions.productDetails(
            product_ids,
        ).call(block_identifier=block_identifier)
        return [
            PredictionProductV1(
                BaseProduct(
                    ProductMetadata(
                        eth_typing.ChecksumAddress(return_value_elem[0][0][0]),
                        str(return_value_elem[0][0][1]),
                        str(return_value_elem[0][0][2]),
                    ),
                    OracleSpecification(
                        eth_typing.ChecksumAddress(return_value_elem[0][1][0]),
                        int(return_value_elem[0][1][1]),
                        int(return_value_elem[0][1][2]),
                        int(return_value_elem[0][1][3]),
                        hexbytes.HexBytes(return_value_elem[0][1][4]),
                    ),
                    eth_typing.ChecksumAddress(return_value_elem[0][2]),
                    int(return_value_elem[0][3]),
                    int(return_value_elem[0][4]),
                    int(return_value_elem[0][5]),
                    str(return_value_elem[0][6]),
                ),
                ExpirySpecification(
                    int(return_value_elem[1][0]), int(return_value_elem[1][1])
                ),
                int(return_value_elem[2]),
                int(return_value_elem[3]),
            )
            for return_value_elem in return_value
        ]

    def product_registry(
        self,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> eth_typing.ChecksumAddress:
        """Binding for `productRegistry` on the SystemViewer contract.

        Parameters
        ----------
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        eth_typing.ChecksumAddress
        """
        return_value = self._contract.functions.productRegistry().call(
            block_identifier=block_identifier
        )
        return eth_typing.ChecksumAddress(return_value)

    def product_states(
        self,
        product_ids: typing.List[hexbytes.HexBytes],
        block_identifier: types.BlockIdentifier = "latest",
    ) -> typing.List[ProductState]:
        """Binding for `productStates` on the SystemViewer contract.

        Parameters
        ----------
        product_ids : typing.List[hexbytes.HexBytes]
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        typing.List[ProductState]
        """
        return_value = self._contract.functions.productStates(
            product_ids,
        ).call(block_identifier=block_identifier)
        return [ProductState(return_value_elem) for return_value_elem in return_value]

    def proxiable_uuid(
        self,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> hexbytes.HexBytes:
        """Binding for `proxiableUUID` on the SystemViewer contract.

        Parameters
        ----------
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        hexbytes.HexBytes
        """
        return_value = self._contract.functions.proxiableUUID().call(
            block_identifier=block_identifier
        )
        return hexbytes.HexBytes(return_value)

    def renounce_ownership(
        self,
    ) -> contract.ContractFunction:
        """Binding for `renounceOwnership` on the SystemViewer contract.

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.renounceOwnership()

    def transfer_ownership(
        self,
        new_owner: eth_typing.ChecksumAddress,
    ) -> contract.ContractFunction:
        """Binding for `transferOwnership` on the SystemViewer contract.

        Parameters
        ----------
        new_owner : eth_typing.ChecksumAddress

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.transferOwnership(
            new_owner,
        )

    def upgrade_to_and_call(
        self,
        new_implementation: eth_typing.ChecksumAddress,
        data: hexbytes.HexBytes,
    ) -> contract.ContractFunction:
        """Binding for `upgradeToAndCall` on the SystemViewer contract.

        Parameters
        ----------
        new_implementation : eth_typing.ChecksumAddress
        data : hexbytes.HexBytes

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.upgradeToAndCall(
            new_implementation,
            data,
        )

    def user_margin_data_by_collateral_asset(
        self,
        collateral_asset: eth_typing.ChecksumAddress,
        accounts: typing.List[eth_typing.ChecksumAddress],
        block_identifier: types.BlockIdentifier = "latest",
    ) -> typing.List[UserMarginAccountData]:
        """Binding for `userMarginDataByCollateralAsset` on the SystemViewer contract.

        Parameters
        ----------
        collateral_asset : eth_typing.ChecksumAddress
        accounts : typing.List[eth_typing.ChecksumAddress]
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        typing.List[UserMarginAccountData]
        """
        return_value = self._contract.functions.userMarginDataByCollateralAsset(
            collateral_asset,
            accounts,
        ).call(block_identifier=block_identifier)
        return [
            UserMarginAccountData(
                eth_typing.ChecksumAddress(return_value_elem[0]),
                eth_typing.ChecksumAddress(return_value_elem[1]),
                int(return_value_elem[2]),
                int(return_value_elem[3]),
                [
                    PositionData(
                        hexbytes.HexBytes(return_value_elem4_elem[0]),
                        int(return_value_elem4_elem[1]),
                        int(return_value_elem4_elem[2]),
                        int(return_value_elem4_elem[3]),
                        int(return_value_elem4_elem[4]),
                    )
                    for return_value_elem4_elem in return_value_elem[4]
                ],
            )
            for return_value_elem in return_value
        ]

    def user_margin_data_by_collateral_assets(
        self,
        collateral_assets: typing.List[eth_typing.ChecksumAddress],
        accounts: typing.List[eth_typing.ChecksumAddress],
        block_identifier: types.BlockIdentifier = "latest",
    ) -> typing.List[UserMarginAccountData]:
        """Binding for `userMarginDataByCollateralAssets` on the SystemViewer contract.

        Parameters
        ----------
        collateral_assets : typing.List[eth_typing.ChecksumAddress]
        accounts : typing.List[eth_typing.ChecksumAddress]
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        typing.List[UserMarginAccountData]
        """
        return_value = self._contract.functions.userMarginDataByCollateralAssets(
            collateral_assets,
            accounts,
        ).call(block_identifier=block_identifier)
        return [
            UserMarginAccountData(
                eth_typing.ChecksumAddress(return_value_elem[0]),
                eth_typing.ChecksumAddress(return_value_elem[1]),
                int(return_value_elem[2]),
                int(return_value_elem[3]),
                [
                    PositionData(
                        hexbytes.HexBytes(return_value_elem4_elem[0]),
                        int(return_value_elem4_elem[1]),
                        int(return_value_elem4_elem[2]),
                        int(return_value_elem4_elem[3]),
                        int(return_value_elem4_elem[4]),
                    )
                    for return_value_elem4_elem in return_value_elem[4]
                ],
            )
            for return_value_elem in return_value
        ]

    def valuations(
        self,
        product_ids: typing.List[hexbytes.HexBytes],
        block_identifier: types.BlockIdentifier = "latest",
    ) -> typing.List[int]:
        """Binding for `valuations` on the SystemViewer contract.

        Parameters
        ----------
        product_ids : typing.List[hexbytes.HexBytes]
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        typing.List[int]
        """
        return_value = self._contract.functions.valuations(
            product_ids,
        ).call(block_identifier=block_identifier)
        return [int(return_value_elem) for return_value_elem in return_value]


ABI = typing.cast(
    eth_typing.ABI,
    [
        {
            "type": "function",
            "name": "UPGRADE_INTERFACE_VERSION",
            "inputs": [],
            "outputs": [{"name": "", "type": "string", "internalType": "string"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "clearing",
            "inputs": [],
            "outputs": [{"name": "", "type": "address", "internalType": "address"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "initialize",
            "inputs": [
                {"name": "_clearing", "type": "address", "internalType": "address"},
                {
                    "name": "_productRegistry",
                    "type": "address",
                    "internalType": "contract IProductRegistry",
                },
                {
                    "name": "_marginAccountRegistry",
                    "type": "address",
                    "internalType": "contract IMarginAccountRegistry",
                },
            ],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "maeByCollateralAsset",
            "inputs": [
                {
                    "name": "collateralAsset",
                    "type": "address",
                    "internalType": "address",
                },
                {"name": "accounts", "type": "address[]", "internalType": "address[]"},
            ],
            "outputs": [
                {"name": "results", "type": "int256[]", "internalType": "int256[]"}
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "maeByCollateralAssets",
            "inputs": [
                {
                    "name": "collateralAssets",
                    "type": "address[]",
                    "internalType": "address[]",
                },
                {"name": "accounts", "type": "address[]", "internalType": "address[]"},
            ],
            "outputs": [
                {"name": "results", "type": "int256[]", "internalType": "int256[]"}
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "maeChecks",
            "inputs": [
                {"name": "accounts", "type": "address[]", "internalType": "address[]"},
                {
                    "name": "settlements",
                    "type": "tuple[]",
                    "internalType": "struct Settlement[]",
                    "components": [
                        {
                            "name": "productId",
                            "type": "bytes32",
                            "internalType": "bytes32",
                        },
                        {
                            "name": "quantity",
                            "type": "int256",
                            "internalType": "int256",
                        },
                        {"name": "price", "type": "int256", "internalType": "int256"},
                    ],
                },
            ],
            "outputs": [
                {"name": "results", "type": "bool[]", "internalType": "bool[]"},
                {"name": "maeAfter", "type": "int256[]", "internalType": "int256[]"},
                {"name": "mmuAfter", "type": "uint256[]", "internalType": "uint256[]"},
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "marginAccountRegistry",
            "inputs": [],
            "outputs": [
                {
                    "name": "",
                    "type": "address",
                    "internalType": "contract IMarginAccountRegistry",
                }
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "owner",
            "inputs": [],
            "outputs": [{"name": "", "type": "address", "internalType": "address"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "positionQuantitiesByProductId",
            "inputs": [
                {"name": "productId", "type": "bytes32", "internalType": "bytes32"},
                {"name": "accounts", "type": "address[]", "internalType": "address[]"},
            ],
            "outputs": [
                {"name": "quantity", "type": "int256[]", "internalType": "int256[]"}
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "positionQuantitiesByProductIds",
            "inputs": [
                {
                    "name": "productIds",
                    "type": "bytes32[]",
                    "internalType": "bytes32[]",
                },
                {"name": "accounts", "type": "address[]", "internalType": "address[]"},
            ],
            "outputs": [
                {"name": "quantity", "type": "int256[]", "internalType": "int256[]"}
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "positionsByCollateralAsset",
            "inputs": [
                {
                    "name": "collateralAsset",
                    "type": "address",
                    "internalType": "address",
                },
                {"name": "accounts", "type": "address[]", "internalType": "address[]"},
            ],
            "outputs": [
                {
                    "name": "results",
                    "type": "tuple[][]",
                    "internalType": "struct PositionData[][]",
                    "components": [
                        {
                            "name": "productId",
                            "type": "bytes32",
                            "internalType": "bytes32",
                        },
                        {
                            "name": "quantity",
                            "type": "int256",
                            "internalType": "int256",
                        },
                        {
                            "name": "costBasis",
                            "type": "int256",
                            "internalType": "int256",
                        },
                        {
                            "name": "maintenanceMargin",
                            "type": "uint256",
                            "internalType": "uint256",
                        },
                        {"name": "pnl", "type": "int256", "internalType": "int256"},
                    ],
                }
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "positionsByCollateralAssets",
            "inputs": [
                {
                    "name": "collateralAssets",
                    "type": "address[]",
                    "internalType": "address[]",
                },
                {"name": "accounts", "type": "address[]", "internalType": "address[]"},
            ],
            "outputs": [
                {
                    "name": "results",
                    "type": "tuple[][]",
                    "internalType": "struct PositionData[][]",
                    "components": [
                        {
                            "name": "productId",
                            "type": "bytes32",
                            "internalType": "bytes32",
                        },
                        {
                            "name": "quantity",
                            "type": "int256",
                            "internalType": "int256",
                        },
                        {
                            "name": "costBasis",
                            "type": "int256",
                            "internalType": "int256",
                        },
                        {
                            "name": "maintenanceMargin",
                            "type": "uint256",
                            "internalType": "uint256",
                        },
                        {"name": "pnl", "type": "int256", "internalType": "int256"},
                    ],
                }
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "productDetails",
            "inputs": [
                {"name": "productIds", "type": "bytes32[]", "internalType": "bytes32[]"}
            ],
            "outputs": [
                {
                    "name": "details",
                    "type": "tuple[]",
                    "internalType": "struct PredictionProductV1[]",
                    "components": [
                        {
                            "name": "base",
                            "type": "tuple",
                            "internalType": "struct BaseProduct",
                            "components": [
                                {
                                    "name": "metadata",
                                    "type": "tuple",
                                    "internalType": "struct ProductMetadata",
                                    "components": [
                                        {
                                            "name": "builder",
                                            "type": "address",
                                            "internalType": "address",
                                        },
                                        {
                                            "name": "symbol",
                                            "type": "string",
                                            "internalType": "string",
                                        },
                                        {
                                            "name": "description",
                                            "type": "string",
                                            "internalType": "string",
                                        },
                                    ],
                                },
                                {
                                    "name": "oracleSpec",
                                    "type": "tuple",
                                    "internalType": "struct OracleSpecification",
                                    "components": [
                                        {
                                            "name": "oracleAddress",
                                            "type": "address",
                                            "internalType": "address",
                                        },
                                        {
                                            "name": "fsvDecimals",
                                            "type": "uint8",
                                            "internalType": "uint8",
                                        },
                                        {
                                            "name": "fspAlpha",
                                            "type": "int256",
                                            "internalType": "int256",
                                        },
                                        {
                                            "name": "fspBeta",
                                            "type": "int256",
                                            "internalType": "int256",
                                        },
                                        {
                                            "name": "fsvCalldata",
                                            "type": "bytes",
                                            "internalType": "bytes",
                                        },
                                    ],
                                },
                                {
                                    "name": "collateralAsset",
                                    "type": "address",
                                    "internalType": "address",
                                },
                                {
                                    "name": "startTime",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                                {
                                    "name": "pointValue",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                                {
                                    "name": "priceDecimals",
                                    "type": "uint8",
                                    "internalType": "uint8",
                                },
                                {
                                    "name": "extendedMetadata",
                                    "type": "string",
                                    "internalType": "string",
                                },
                            ],
                        },
                        {
                            "name": "expirySpec",
                            "type": "tuple",
                            "internalType": "struct ExpirySpecification",
                            "components": [
                                {
                                    "name": "earliestFSPSubmissionTime",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                                {
                                    "name": "tradeoutInterval",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                            ],
                        },
                        {
                            "name": "maxPrice",
                            "type": "int256",
                            "internalType": "int256",
                        },
                        {
                            "name": "minPrice",
                            "type": "int256",
                            "internalType": "int256",
                        },
                    ],
                }
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "productRegistry",
            "inputs": [],
            "outputs": [
                {
                    "name": "",
                    "type": "address",
                    "internalType": "contract IProductRegistry",
                }
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "productStates",
            "inputs": [
                {"name": "productIds", "type": "bytes32[]", "internalType": "bytes32[]"}
            ],
            "outputs": [
                {
                    "name": "states",
                    "type": "uint8[]",
                    "internalType": "enum ProductState[]",
                }
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "proxiableUUID",
            "inputs": [],
            "outputs": [{"name": "", "type": "bytes32", "internalType": "bytes32"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "renounceOwnership",
            "inputs": [],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "transferOwnership",
            "inputs": [
                {"name": "newOwner", "type": "address", "internalType": "address"}
            ],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "function",
            "name": "upgradeToAndCall",
            "inputs": [
                {
                    "name": "newImplementation",
                    "type": "address",
                    "internalType": "address",
                },
                {"name": "data", "type": "bytes", "internalType": "bytes"},
            ],
            "outputs": [],
            "stateMutability": "payable",
        },
        {
            "type": "function",
            "name": "userMarginDataByCollateralAsset",
            "inputs": [
                {
                    "name": "collateralAsset",
                    "type": "address",
                    "internalType": "address",
                },
                {"name": "accounts", "type": "address[]", "internalType": "address[]"},
            ],
            "outputs": [
                {
                    "name": "data",
                    "type": "tuple[]",
                    "internalType": "struct ISystemViewer.UserMarginAccountData[]",
                    "components": [
                        {
                            "name": "collateralAsset",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "marginAccountId",
                            "type": "address",
                            "internalType": "address",
                        },
                        {"name": "mae", "type": "int256", "internalType": "int256"},
                        {"name": "mmu", "type": "uint256", "internalType": "uint256"},
                        {
                            "name": "positions",
                            "type": "tuple[]",
                            "internalType": "struct PositionData[]",
                            "components": [
                                {
                                    "name": "productId",
                                    "type": "bytes32",
                                    "internalType": "bytes32",
                                },
                                {
                                    "name": "quantity",
                                    "type": "int256",
                                    "internalType": "int256",
                                },
                                {
                                    "name": "costBasis",
                                    "type": "int256",
                                    "internalType": "int256",
                                },
                                {
                                    "name": "maintenanceMargin",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                                {
                                    "name": "pnl",
                                    "type": "int256",
                                    "internalType": "int256",
                                },
                            ],
                        },
                    ],
                }
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "userMarginDataByCollateralAssets",
            "inputs": [
                {
                    "name": "collateralAssets",
                    "type": "address[]",
                    "internalType": "address[]",
                },
                {"name": "accounts", "type": "address[]", "internalType": "address[]"},
            ],
            "outputs": [
                {
                    "name": "data",
                    "type": "tuple[]",
                    "internalType": "struct ISystemViewer.UserMarginAccountData[]",
                    "components": [
                        {
                            "name": "collateralAsset",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "marginAccountId",
                            "type": "address",
                            "internalType": "address",
                        },
                        {"name": "mae", "type": "int256", "internalType": "int256"},
                        {"name": "mmu", "type": "uint256", "internalType": "uint256"},
                        {
                            "name": "positions",
                            "type": "tuple[]",
                            "internalType": "struct PositionData[]",
                            "components": [
                                {
                                    "name": "productId",
                                    "type": "bytes32",
                                    "internalType": "bytes32",
                                },
                                {
                                    "name": "quantity",
                                    "type": "int256",
                                    "internalType": "int256",
                                },
                                {
                                    "name": "costBasis",
                                    "type": "int256",
                                    "internalType": "int256",
                                },
                                {
                                    "name": "maintenanceMargin",
                                    "type": "uint256",
                                    "internalType": "uint256",
                                },
                                {
                                    "name": "pnl",
                                    "type": "int256",
                                    "internalType": "int256",
                                },
                            ],
                        },
                    ],
                }
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "valuations",
            "inputs": [
                {"name": "productIds", "type": "bytes32[]", "internalType": "bytes32[]"}
            ],
            "outputs": [
                {"name": "prices", "type": "int256[]", "internalType": "int256[]"}
            ],
            "stateMutability": "view",
        },
        {
            "type": "event",
            "name": "Initialized",
            "inputs": [
                {
                    "name": "version",
                    "type": "uint64",
                    "indexed": False,
                    "internalType": "uint64",
                }
            ],
            "anonymous": False,
        },
        {
            "type": "event",
            "name": "OwnershipTransferred",
            "inputs": [
                {
                    "name": "previousOwner",
                    "type": "address",
                    "indexed": True,
                    "internalType": "address",
                },
                {
                    "name": "newOwner",
                    "type": "address",
                    "indexed": True,
                    "internalType": "address",
                },
            ],
            "anonymous": False,
        },
        {
            "type": "event",
            "name": "Upgraded",
            "inputs": [
                {
                    "name": "implementation",
                    "type": "address",
                    "indexed": True,
                    "internalType": "address",
                }
            ],
            "anonymous": False,
        },
        {
            "type": "error",
            "name": "AddressEmptyCode",
            "inputs": [
                {"name": "target", "type": "address", "internalType": "address"}
            ],
        },
        {
            "type": "error",
            "name": "ERC1967InvalidImplementation",
            "inputs": [
                {"name": "implementation", "type": "address", "internalType": "address"}
            ],
        },
        {"type": "error", "name": "ERC1967NonPayable", "inputs": []},
        {"type": "error", "name": "FailedCall", "inputs": []},
        {"type": "error", "name": "InvalidInitialization", "inputs": []},
        {
            "type": "error",
            "name": "InvalidParameters",
            "inputs": [{"name": "reason", "type": "string", "internalType": "string"}],
        },
        {"type": "error", "name": "NotInitializing", "inputs": []},
        {
            "type": "error",
            "name": "OwnableInvalidOwner",
            "inputs": [{"name": "owner", "type": "address", "internalType": "address"}],
        },
        {
            "type": "error",
            "name": "OwnableUnauthorizedAccount",
            "inputs": [
                {"name": "account", "type": "address", "internalType": "address"}
            ],
        },
        {"type": "error", "name": "UUPSUnauthorizedCallContext", "inputs": []},
        {
            "type": "error",
            "name": "UUPSUnsupportedProxiableUUID",
            "inputs": [{"name": "slot", "type": "bytes32", "internalType": "bytes32"}],
        },
    ],
)



from pydantic import ValidationError

from afp.dtos import PaginationFilter
from afp.schemas import AliasedModel


class Person(AliasedModel):
    first_name: str
    last_name: str


def test_schema_immutability():
    person = Person(first_name="Foo", last_name="Bar")
    try:
        person.first_name = "Baz"
    except ValidationError:
        pass
    else:
        assert False


def test_schema_aliasing__from_json():
    person = Person.model_validate_json('{"firstName":"Foo","lastName":"Bar"}')
    assert person.first_name == "Foo"
    assert person.last_name == "Bar"


def test_schema_aliasing__to_json():
    person = Person(first_name="Foo", last_name="Bar")
    assert person.model_dump_json() == '{"firstName":"Foo","lastName":"Bar"}'


def test_pagination_parameters__conversion():
    filter = PaginationFilter(batch=2, batch_size=20, newest_first=False)
    assert filter.model_dump() == {
        "page": 2,
        "page_size": 20,
        "sort": "ASC",
    }


def test_pagination_parameters__null_values():
    filter = PaginationFilter(batch=None, batch_size=None, newest_first=None)
    assert filter.model_dump(exclude_none=True) == {}



from abc import ABC
from datetime import datetime
from functools import cache
from urllib.parse import urlparse
from typing import cast

from eth_typing.evm import ChecksumAddress
from siwe import ISO8601Datetime, SiweMessage, siwe  # type: ignore (untyped library)
from web3 import Web3, HTTPProvider
from web3.contract.contract import ContractFunction
from web3.types import TxParams

from ..auth import Authenticator
from ..bindings.erc20 import ERC20
from ..config import Config
from ..dtos import LoginSubmission
from ..exceptions import ConfigurationError
from ..exchange import ExchangeClient
from ..ipfs import IPFSClient
from ..schemas import Transaction


class BaseAPI(ABC):
    _authenticator: Authenticator
    _config: Config

    def __init__(self, config: Config, authenticator: Authenticator | None = None):
        self._config = config

        if authenticator is None:
            if config.authenticator is None:
                raise ConfigurationError("Authenticator not specified")
            self._authenticator = config.authenticator
        else:
            self._authenticator = authenticator

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(authenticator={repr(self._authenticator)})"


class ClearingSystemAPI(BaseAPI, ABC):
    _config: Config
    _w3: Web3

    def __init__(self, config: Config, authenticator: Authenticator | None = None):
        super().__init__(config, authenticator)

        if self._config.rpc_url is None:
            raise ConfigurationError("RPC URL not specified")

        self._w3 = Web3(HTTPProvider(self._config.rpc_url))
        self._w3.eth.default_account = self._authenticator.address

    def _transact(self, func: ContractFunction) -> Transaction:
        tx_count = self._w3.eth.get_transaction_count(self._authenticator.address)
        tx_params = {
            "from": self._authenticator.address,
            "nonce": tx_count,
            "gasLimit": self._config.gas_limit,
            "maxFeePerGas": self._config.max_fee_per_gas,
            "maxPriorityFeePerGas": self._config.max_priority_fee_per_gas,
        }

        prepared_tx = func.build_transaction(
            cast(TxParams, {k: v for k, v in tx_params.items() if v is not None})
        )
        signed_tx = self._authenticator.sign_transaction(prepared_tx)
        tx_hash = self._w3.eth.send_raw_transaction(signed_tx.raw_transaction)
        tx_receipt = self._w3.eth.wait_for_transaction_receipt(
            tx_hash, timeout=self._config.timeout_seconds
        )

        return Transaction(
            hash=tx_hash.to_0x_hex(),
            data=dict(prepared_tx),
            receipt=dict(tx_receipt),
        )

    @cache
    def _decimals(self, collateral_asset: ChecksumAddress) -> int:
        token_contract = ERC20(self._w3, collateral_asset)
        return token_contract.decimals()


class ExchangeAPI(BaseAPI, ABC):
    _exchange: ExchangeClient
    _trading_protocol_id: str

    def __init__(
        self,
        config: Config,
        authenticator: Authenticator | None = None,
        exchange_url: str | None = None,
    ):
        if exchange_url is None:
            exchange_url = config.exchange_url

        super().__init__(config, authenticator)
        self._exchange = ExchangeClient(exchange_url)
        self._login()

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(authenticator={repr(self._authenticator)}, "
            f"exchange={repr(self._exchange)})"
        )

    def _login(self):
        nonce = self._exchange.generate_login_nonce()
        message = self._generate_eip4361_message(nonce)
        signature = self._authenticator.sign_message(message.encode("ascii"))

        login_submission = LoginSubmission(
            message=message, signature=Web3.to_hex(signature)
        )
        exchange_parameters = self._exchange.login(login_submission)

        self._trading_protocol_id = exchange_parameters.trading_protocol_id

    def _generate_eip4361_message(self, nonce: str) -> str:
        message = SiweMessage(
            domain=urlparse(self._config.exchange_url).netloc,
            address=self._authenticator.address,
            uri=self._config.exchange_url,
            version=siwe.VersionEnum.one,  # type: ignore
            chain_id=self._config.chain_id,
            issued_at=ISO8601Datetime.from_datetime(datetime.now()),
            nonce=nonce,
            statement=None,
        )
        return message.prepare_message()


class IPFSManager(ABC):
    _ipfs_client: IPFSClient

    def __init__(self, config: Config):
        self._ipfs_client = IPFSClient(config.ipfs_api_url, config.ipfs_api_key)



"""Autonomous Futures Protocol Python SDK."""

from . import bindings, enums, exceptions, schemas
from .afp import AFP
from .auth import (
    Authenticator,
    KeyfileAuthenticator,
    PrivateKeyAuthenticator,
    TrezorAuthenticator,
)
from .exceptions import AFPException

__all__ = (
    "bindings",
    "enums",
    "exceptions",
    "schemas",
    "AFP",
    "AFPException",
    "Authenticator",
    "KeyfileAuthenticator",
    "PrivateKeyAuthenticator",
    "TrezorAuthenticator",
)



class AFPException(Exception):
    pass


# API-specific exceptions


class ConfigurationError(AFPException):
    pass


class ClearingSystemError(AFPException):
    pass


class IPFSError(AFPException):
    pass


class ExchangeError(AFPException):
    pass


class DeviceError(AFPException):
    pass


# Exchange error sub-types


class AuthenticationError(ExchangeError):
    pass


class AuthorizationError(ExchangeError):
    pass


class NotFoundError(ExchangeError):
    pass


class RateLimitExceeded(ExchangeError):
    pass


class ValidationError(ExchangeError):
    pass



from datetime import datetime
from unittest.mock import Mock

import pytest
import requests
from requests import Response
from requests.adapters import HTTPAdapter

from afp.dtos import (
    ExchangeProductFilter,
    OrderFillFilter,
    OrderFilter,
)
from afp.enums import OrderSide, OrderState, TradeState
from afp.exceptions import (
    AuthenticationError,
    AuthorizationError,
    ExchangeError,
    NotFoundError,
    RateLimitExceeded,
    ValidationError,
)
from afp.exchange import ExchangeClient

from .fixtures import (
    make_error_response,
    make_exchange_parameters,
    make_exchange_product,
    make_login_submission,
    make_market_depth_data,
    make_ndjson_response,
    make_ohlcv_item,
    make_order,
    make_order_fill,
    make_order_submission,
    make_product_listing_submission,
    make_product_update_submission,
    make_success_response,
)


def test_generate_login_nonce__success__returns_message(monkeypatch):
    """Test successful nonce generation."""
    fake_response = make_success_response({"message": "test-nonce-12345"})
    mock_send = Mock(return_value=fake_response)
    monkeypatch.setattr(HTTPAdapter, "send", mock_send)

    client = ExchangeClient("http://test.com")
    nonce = client.generate_login_nonce()

    assert nonce == "test-nonce-12345"
    request = mock_send.call_args[0][0]
    assert request.method == "GET"
    assert "v1/nonce" in request.url


def test_login__success__returns_exchange_parameters(monkeypatch):
    """Test successful login with signature."""
    params = make_exchange_parameters(
        trading_protocol_id="0xabcd",
        maker_trading_fee_rate="0.001",
        taker_trading_fee_rate="0.002",
    )
    fake_response = make_success_response(params.model_dump())
    mock_send = Mock(return_value=fake_response)
    monkeypatch.setattr(HTTPAdapter, "send", mock_send)

    client = ExchangeClient("http://test.com")
    login_submission = make_login_submission(
        message="test-nonce",
        signature="0xsignature",
    )
    result = client.login(login_submission)

    assert result.trading_protocol_id == "0xabcd"
    assert str(result.maker_trading_fee_rate) == "0.001"

    request = mock_send.call_args[0][0]
    assert request.method == "POST"
    assert "v1/login" in request.url
    assert "test-nonce" in request.body


def test_login__sends_correct_headers(monkeypatch):
    """Test that login request includes correct headers."""
    params = make_exchange_parameters()
    fake_response = make_success_response(params.model_dump())
    mock_send = Mock(return_value=fake_response)
    monkeypatch.setattr(HTTPAdapter, "send", mock_send)

    client = ExchangeClient("http://test.com")
    client.login(make_login_submission())

    request = mock_send.call_args[0][0]
    assert request.headers["Content-Type"] == "application/json"
    assert request.headers["Accept"] == "application/json"
    assert "afp-sdk/" in request.headers["User-Agent"]


def test_get_approved_products__success__returns_list(monkeypatch):
    """Test successful retrieval of approved products."""
    product1 = make_exchange_product(id="prod1", symbol="BTCUSD")
    product2 = make_exchange_product(id="prod2", symbol="ETHUSD")
    fake_response = make_success_response(
        {"products": [product1.model_dump(), product2.model_dump()]}
    )
    mock_send = Mock(return_value=fake_response)
    monkeypatch.setattr(HTTPAdapter, "send", mock_send)

    client = ExchangeClient("http://test.com")
    filter_obj = ExchangeProductFilter(
        batch=None,
        batch_size=None,
        newest_first=None,
    )
    products = client.get_approved_products(filter_obj)

    assert len(products) == 2
    assert products[0].id == "prod1"
    assert products[0].symbol == "BTCUSD"
    assert products[1].id == "prod2"

    request = mock_send.call_args[0][0]
    assert request.method == "GET"
    assert "v1/products" in request.url


def test_get_approved_products__with_pagination__includes_params(monkeypatch):
    """Test that pagination parameters are included in request."""
    fake_response = make_success_response({"products": []})
    mock_send = Mock(return_value=fake_response)
    monkeypatch.setattr(HTTPAdapter, "send", mock_send)

    client = ExchangeClient("http://test.com")
    filter_obj = ExchangeProductFilter(
        batch=2,
        batch_size=50,
        newest_first=True,
    )
    client.get_approved_products(filter_obj)

    request = mock_send.call_args[0][0]
    assert "page=2" in request.url
    assert "page_size=50" in request.url
    assert "sort=DESC" in request.url


def test_get_product_by_id__success__returns_product(monkeypatch):
    """Test successful retrieval of single product by ID."""
    product = make_exchange_product(id="prod123", symbol="BTCUSD")
    fake_response = make_success_response(product.model_dump())
    mock_send = Mock(return_value=fake_response)
    monkeypatch.setattr(HTTPAdapter, "send", mock_send)

    client = ExchangeClient("http://test.com")
    result = client.get_product_by_id("prod123")

    assert result.id == "prod123"
    assert result.symbol == "BTCUSD"

    request = mock_send.call_args[0][0]
    assert request.method == "GET"
    assert "v1/products/prod123" in request.url


def test_list_product__success__returns_none(monkeypatch):
    """Test successful product listing submission."""
    fake_response = Response()
    fake_response.status_code = 200
    fake_response._content = b""  # type: ignore
    fake_response.history = []  # type: ignore
    mock_send = Mock(return_value=fake_response)
    monkeypatch.setattr(HTTPAdapter, "send", mock_send)

    client = ExchangeClient("http://test.com")
    listing = make_product_listing_submission()
    result = client.list_product(listing)

    assert result is None

    request = mock_send.call_args[0][0]
    assert request.method == "POST"
    assert "v1/products" in request.url
    assert listing.id in request.body


def test_update_product_listing__success__returns_none(monkeypatch):
    """Test successful product listing update."""
    fake_response = Response()
    fake_response.status_code = 200
    fake_response._content = b""  # type: ignore
    fake_response.history = []  # type: ignore
    mock_send = Mock(return_value=fake_response)
    monkeypatch.setattr(HTTPAdapter, "send", mock_send)

    client = ExchangeClient("http://test.com")
    update = make_product_update_submission()
    result = client.update_product_listing("prod123", update)

    assert result is None

    request = mock_send.call_args[0][0]
    assert request.method == "PATCH"
    assert "v1/products/prod123" in request.url


def test_submit_order__success__returns_order(monkeypatch):
    """Test successful order submission."""
    order = make_order(id="order123", state=OrderState.OPEN)
    fake_response = make_success_response(order.model_dump())
    mock_send = Mock(return_value=fake_response)
    monkeypatch.setattr(HTTPAdapter, "send", mock_send)

    client = ExchangeClient("http://test.com")
    submission = make_order_submission()
    result = client.submit_order(submission)

    assert result.id == "order123"
    assert result.state == OrderState.OPEN

    request = mock_send.call_args[0][0]
    assert request.method == "POST"
    assert "v1/orders" in request.url


def test_get_orders__success__returns_list(monkeypatch):
    """Test successful retrieval of orders."""
    order1 = make_order(id="order1")
    order2 = make_order(id="order2")
    fake_response = make_success_response(
        {"orders": [order1.model_dump(), order2.model_dump()]}
    )
    mock_send = Mock(return_value=fake_response)
    monkeypatch.setattr(HTTPAdapter, "send", mock_send)

    client = ExchangeClient("http://test.com")
    filter_obj = OrderFilter(
        intent_account_id="0x1111111111111111111111111111111111111111",
        product_id=None,
        type=None,
        states=[OrderState.OPEN],
        side=OrderSide.BID,
        start=None,
        end=None,
        batch=None,
        batch_size=None,
        newest_first=None,
    )
    orders = client.get_orders(filter_obj)

    assert len(orders) == 2
    assert orders[0].id == "order1"

    request = mock_send.call_args[0][0]
    assert request.method == "GET"
    assert "v1/orders" in request.url
    assert "side=BID" in request.url


def test_get_orders__with_state_filter__includes_comma_separated_states(monkeypatch):
    """Test that multiple states are comma-separated in query params."""
    fake_response = make_success_response({"orders": []})
    mock_send = Mock(return_value=fake_response)
    monkeypatch.setattr(HTTPAdapter, "send", mock_send)

    client = ExchangeClient("http://test.com")
    filter_obj = OrderFilter(
        intent_account_id="0x1111111111111111111111111111111111111111",
        product_id=None,
        type=None,
        states=[OrderState.OPEN, OrderState.COMPLETED],
        side=None,
        start=None,
        end=None,
        batch=None,
        batch_size=None,
        newest_first=None,
    )
    client.get_orders(filter_obj)

    request = mock_send.call_args[0][0]
    assert (
        "state=OPEN%2CCOMPLETED" in request.url or "state=OPEN,COMPLETED" in request.url
    )


def test_get_order_by_id__success__returns_order(monkeypatch):
    """Test successful retrieval of single order by ID."""
    order = make_order(id="order123")
    fake_response = make_success_response(order.model_dump())
    mock_send = Mock(return_value=fake_response)
    monkeypatch.setattr(HTTPAdapter, "send", mock_send)

    client = ExchangeClient("http://test.com")
    result = client.get_order_by_id("order123")

    assert result.id == "order123"

    request = mock_send.call_args[0][0]
    assert request.method == "GET"
    assert "v1/orders/order123" in request.url


def test_get_order_fills__success__returns_list(monkeypatch):
    """Test successful retrieval of order fills."""
    fill1 = make_order_fill(quantity=10)
    fill2 = make_order_fill(quantity=20)
    fake_response = make_success_response(
        {"orderFills": [fill1.model_dump(), fill2.model_dump()]}
    )
    mock_send = Mock(return_value=fake_response)
    monkeypatch.setattr(HTTPAdapter, "send", mock_send)

    client = ExchangeClient("http://test.com")
    filter_obj = OrderFillFilter(
        intent_account_id="0x1111111111111111111111111111111111111111",
        product_id=None,
        intent_hash=None,
        start=None,
        end=None,
        trade_states=[TradeState.CLEARED],
        batch=None,
        batch_size=None,
        newest_first=None,
    )
    fills = client.get_order_fills(filter_obj)

    assert len(fills) == 2
    assert fills[0].quantity == 10
    assert fills[1].quantity == 20

    request = mock_send.call_args[0][0]
    assert request.method == "GET"
    assert "v1/order-fills" in request.url


def test_iter_order_fills__success__yields_items(monkeypatch):
    """Test NDJSON streaming for order fills."""
    fill1 = make_order_fill(quantity=10)
    fill2 = make_order_fill(quantity=20)
    fake_response = make_ndjson_response([fill1, fill2])
    mock_send = Mock(return_value=fake_response)
    monkeypatch.setattr(HTTPAdapter, "send", mock_send)

    client = ExchangeClient("http://test.com")
    filter_obj = OrderFillFilter(
        intent_account_id="0x1111111111111111111111111111111111111111",
        product_id=None,
        intent_hash=None,
        start=None,
        end=None,
        trade_states=[],
        batch=None,
        batch_size=None,
        newest_first=None,
    )
    fills = list(client.iter_order_fills(filter_obj))

    assert len(fills) == 2
    assert fills[0].quantity == 10
    assert fills[1].quantity == 20

    request = mock_send.call_args[0][0]
    assert request.method == "GET"
    assert "v1/stream/order-fills" in request.url
    assert request.headers["Accept"] == "application/x-ndjson"


def test_get_market_depth_data__success__returns_depth(monkeypatch):
    """Test successful retrieval of market depth data."""
    depth_data = make_market_depth_data(product_id="prod123")
    fake_response = make_success_response(depth_data.model_dump())
    mock_send = Mock(return_value=fake_response)
    monkeypatch.setattr(HTTPAdapter, "send", mock_send)

    client = ExchangeClient("http://test.com")
    result = client.get_market_depth_data("prod123")

    assert result.product_id == "prod123"
    assert len(result.bids) == 2
    assert len(result.asks) == 2

    request = mock_send.call_args[0][0]
    assert request.method == "GET"
    assert "v1/market-depth/prod123" in request.url


def test_iter_market_depth_data__success__yields_snapshots(monkeypatch):
    """Test NDJSON streaming for market depth data."""
    depth1 = make_market_depth_data(product_id="prod123")
    depth2 = make_market_depth_data(product_id="prod123")
    fake_response = make_ndjson_response([depth1, depth2])
    mock_send = Mock(return_value=fake_response)
    monkeypatch.setattr(HTTPAdapter, "send", mock_send)

    client = ExchangeClient("http://test.com")
    snapshots = list(client.iter_market_depth_data("prod123"))

    assert len(snapshots) == 2
    assert all(s.product_id == "prod123" for s in snapshots)

    request = mock_send.call_args[0][0]
    assert request.method == "GET"
    assert "v1/stream/market-depth/prod123" in request.url


def test_get_time_series_data__success__returns_ohlcv_list(monkeypatch):
    """Test successful retrieval of time series data."""
    item1 = make_ohlcv_item(timestamp=1700000000)
    item2 = make_ohlcv_item(timestamp=1700000060)
    fake_response = make_success_response(
        {"data": [item1.model_dump(), item2.model_dump()]}
    )
    mock_send = Mock(return_value=fake_response)
    monkeypatch.setattr(HTTPAdapter, "send", mock_send)

    client = ExchangeClient("http://test.com")
    result = client.get_time_series_data("prod123", start=1700000000, interval=60)

    assert len(result) == 2
    # Timestamps are converted to datetime objects
    assert isinstance(result[0].timestamp, datetime)
    assert isinstance(result[1].timestamp, datetime)

    request = mock_send.call_args[0][0]
    assert request.method == "GET"
    assert "v1/time-series/prod123" in request.url
    assert "start=1700000000" in request.url
    assert "interval=60" in request.url


def test_iter_time_series_data__success__yields_items(monkeypatch):
    """Test NDJSON streaming for time series data."""
    item1 = make_ohlcv_item(timestamp=1700000000)
    item2 = make_ohlcv_item(timestamp=1700000060)
    fake_response = make_ndjson_response([item1, item2])
    mock_send = Mock(return_value=fake_response)
    monkeypatch.setattr(HTTPAdapter, "send", mock_send)

    client = ExchangeClient("http://test.com")
    items = list(client.iter_time_series_data("prod123", start=1700000000, interval=60))

    assert len(items) == 2
    # Timestamps are converted to datetime objects
    assert isinstance(items[0].timestamp, datetime)

    request = mock_send.call_args[0][0]
    assert request.method == "GET"
    assert "v1/stream/time-series/prod123" in request.url


def test_send_request__401__raises_authentication_error(monkeypatch):
    """Test that 401 status code raises AuthenticationError."""
    fake_response = make_error_response(401, "Invalid credentials")
    mock_send = Mock(return_value=fake_response)
    monkeypatch.setattr(HTTPAdapter, "send", mock_send)

    client = ExchangeClient("http://test.com")
    with pytest.raises(AuthenticationError):
        client.generate_login_nonce()


def test_send_request__403__raises_authorization_error(monkeypatch):
    """Test that 403 status code raises AuthorizationError."""
    fake_response = make_error_response(403, "Forbidden")
    mock_send = Mock(return_value=fake_response)
    monkeypatch.setattr(HTTPAdapter, "send", mock_send)

    client = ExchangeClient("http://test.com")
    with pytest.raises(AuthorizationError):
        client.generate_login_nonce()


def test_send_request__404__raises_not_found_error(monkeypatch):
    """Test that 404 status code raises NotFoundError."""
    fake_response = make_error_response(404, "Resource not found")
    mock_send = Mock(return_value=fake_response)
    monkeypatch.setattr(HTTPAdapter, "send", mock_send)

    client = ExchangeClient("http://test.com")
    with pytest.raises(NotFoundError):
        client.get_product_by_id("nonexistent")


def test_send_request__429__raises_rate_limit_exceeded(monkeypatch):
    """Test that 429 status code raises RateLimitExceeded."""
    fake_response = make_error_response(429, "Too many requests")
    mock_send = Mock(return_value=fake_response)
    monkeypatch.setattr(HTTPAdapter, "send", mock_send)

    client = ExchangeClient("http://test.com")
    with pytest.raises(RateLimitExceeded):
        client.generate_login_nonce()


def test_send_request__400_with_detail__raises_validation_error(monkeypatch):
    """Test that 400 with detail raises ValidationError."""
    fake_response = make_error_response(400, "Invalid quantity")
    mock_send = Mock(return_value=fake_response)
    monkeypatch.setattr(HTTPAdapter, "send", mock_send)

    client = ExchangeClient("http://test.com")
    with pytest.raises(ValidationError, match="Invalid quantity"):
        client.submit_order(make_order_submission())


def test_send_request__422_with_detail_list__raises_validation_error(monkeypatch):
    """Test that 422 with detail list raises ValidationError."""
    detail = [
        {"msg": "Field is required", "loc": ["body", "quantity"]},
        {"msg": "Invalid type", "loc": ["body", "price"]},
    ]
    fake_response = make_error_response(422, detail)
    mock_send = Mock(return_value=fake_response)
    monkeypatch.setattr(HTTPAdapter, "send", mock_send)

    client = ExchangeClient("http://test.com")
    with pytest.raises(ValidationError, match="Field is required"):
        client.submit_order(make_order_submission())


def test_send_request__500__raises_exchange_error(monkeypatch):
    """Test that 500 status code raises ExchangeError."""
    fake_response = make_error_response(500, "Internal server error")
    mock_send = Mock(return_value=fake_response)
    monkeypatch.setattr(HTTPAdapter, "send", mock_send)

    client = ExchangeClient("http://test.com")
    with pytest.raises(ExchangeError):
        client.generate_login_nonce()


def test_send_request__network_timeout__raises_exchange_error(monkeypatch):
    """Test that network timeout raises ExchangeError."""
    mock_send = Mock(side_effect=requests.exceptions.Timeout("Connection timeout"))
    monkeypatch.setattr(HTTPAdapter, "send", mock_send)

    client = ExchangeClient("http://test.com")
    with pytest.raises(ExchangeError, match="Failed to send request"):
        client.generate_login_nonce()


def test_send_request__connection_error__raises_exchange_error(monkeypatch):
    """Test that connection error raises ExchangeError."""
    mock_send = Mock(
        side_effect=requests.exceptions.ConnectionError("Connection refused")
    )
    monkeypatch.setattr(HTTPAdapter, "send", mock_send)

    client = ExchangeClient("http://test.com")
    with pytest.raises(ExchangeError, match="Failed to send request"):
        client.generate_login_nonce()


def test_send_request__constructs_correct_url(monkeypatch):
    """Test that URLs are constructed correctly."""
    fake_response = make_success_response({"message": "test"})
    mock_send = Mock(return_value=fake_response)
    monkeypatch.setattr(HTTPAdapter, "send", mock_send)

    client = ExchangeClient("http://test.com/")
    client._send_request("GET", "/nonce", api_version=1)

    request = mock_send.call_args[0][0]
    assert request.url == "http://test.com/v1/nonce"


def test_send_request__with_custom_api_version__uses_correct_version(monkeypatch):
    """Test that custom API version is used in URL."""
    fake_response = make_success_response({"message": "test"})
    mock_send = Mock(return_value=fake_response)
    monkeypatch.setattr(HTTPAdapter, "send", mock_send)

    client = ExchangeClient("http://test.com")
    client._send_request("GET", "/nonce", api_version=2)

    request = mock_send.call_args[0][0]
    assert "v2/nonce" in request.url


def test_send_request__streaming_sets_ndjson_accept_header(monkeypatch):
    """Test that streaming requests set NDJSON accept header."""
    fake_response = make_ndjson_response([])
    mock_send = Mock(return_value=fake_response)
    monkeypatch.setattr(HTTPAdapter, "send", mock_send)

    client = ExchangeClient("http://test.com")
    filter_obj = OrderFillFilter(
        intent_account_id="0x1111111111111111111111111111111111111111",
        product_id=None,
        intent_hash=None,
        start=None,
        end=None,
        trade_states=[],
        batch=None,
        batch_size=None,
        newest_first=None,
    )
    list(client.iter_order_fills(filter_obj))

    request = mock_send.call_args[0][0]
    assert request.headers["Accept"] == "application/x-ndjson"


def test_send_request__non_streaming_sets_json_accept_header(monkeypatch):
    """Test that non-streaming requests set JSON accept header."""
    fake_response = make_success_response({"message": "test"})
    mock_send = Mock(return_value=fake_response)
    monkeypatch.setattr(HTTPAdapter, "send", mock_send)

    client = ExchangeClient("http://test.com")
    client.generate_login_nonce()

    request = mock_send.call_args[0][0]
    assert request.headers["Accept"] == "application/json"


def test_send_request__always_sets_content_type_header(monkeypatch):
    """Test that all requests set Content-Type header."""
    fake_response = make_success_response({"message": "test"})
    mock_send = Mock(return_value=fake_response)
    monkeypatch.setattr(HTTPAdapter, "send", mock_send)

    client = ExchangeClient("http://test.com")
    client.generate_login_nonce()

    request = mock_send.call_args[0][0]
    assert request.headers["Content-Type"] == "application/json"



from datetime import datetime, timezone
from decimal import Decimal

from web3 import Web3

from afp.schemas import PredictionProductV1


def test_product_parsing_from_dictionary():
    spec = {
        "base": {
            "metadata": {
                "builder": "0xFfbf2643CF22760AfD3b878BA8aE849c48944Aa5",
                "symbol": "BTCUSDPERP",
                "description": "Bitcoin perpetual futures",
            },
            "oracleSpec": {
                "oracleAddress": "0x1234567890123456789012345678901234567890",
                "fsvDecimals": 18,
                "fspAlpha": "1.0",
                "fspBeta": "0.5",
                "fsvCalldata": "0x1234",
            },
            "collateralAsset": "0xAbCdEf1234567890AbCdEf1234567890AbCdEf12",
            "startTime": "2024-01-01T00:00Z",
            "pointValue": "1.0",
            "priceDecimals": 2,
            "extendedMetadata": "QmfXNPrsPvhKmMPKxsxr6Z3TzH35eHykzpQrruSjPFqeYa",
        },
        "expirySpec": {
            "earliestFSPSubmissionTime": "2024-01-01T12:00Z",
            "tradeoutInterval": 3600,
        },
        "minPrice": "0.01",
        "maxPrice": "99.99",
    }

    result = PredictionProductV1.model_validate(spec)

    assert isinstance(result, PredictionProductV1)

    assert result.base.metadata.builder == "0xFfbf2643CF22760AfD3b878BA8aE849c48944Aa5"
    assert result.base.metadata.symbol == "BTCUSDPERP"
    assert result.base.metadata.description == "Bitcoin perpetual futures"

    assert (
        result.base.oracle_spec.oracle_address
        == "0x1234567890123456789012345678901234567890"
    )
    assert result.base.oracle_spec.fsv_decimals == 18
    assert result.base.oracle_spec.fsp_alpha == Decimal("1.0")
    assert result.base.oracle_spec.fsp_beta == Decimal("0.5")
    assert result.base.oracle_spec.fsv_calldata == "0x1234"

    assert result.base.collateral_asset == Web3.to_checksum_address(
        "0xAbCdEf1234567890AbCdEf1234567890AbCdEf12"
    )
    assert result.base.start_time == datetime(2024, 1, 1, 0, 0, tzinfo=timezone.utc)
    assert result.base.point_value == Decimal("1.0")
    assert result.base.price_decimals == 2
    assert (
        result.base.extended_metadata
        == "QmfXNPrsPvhKmMPKxsxr6Z3TzH35eHykzpQrruSjPFqeYa"
    )

    assert result.expiry_spec.earliest_fsp_submission_time == datetime(
        2024, 1, 1, 12, 0, tzinfo=timezone.utc
    )
    assert result.expiry_spec.tradeout_interval == 3600

    assert result.min_price == Decimal("0.01")
    assert result.max_price == Decimal("99.99")



from datetime import datetime
from decimal import Decimal
from typing import Any, cast

from eth_typing.evm import ChecksumAddress
from hexbytes import HexBytes
from web3 import Web3

from .. import constants, hashing, validators
from ..auth import Authenticator
from ..config import Config
from ..bindings import (
    BaseProduct as OnChainBaseProduct,
    ClearingDiamond,
    ExpirySpecification as OnChainExpirySpecification,
    OracleSpecification as OnChainOracleSpecification,
    PredictionProductV1 as OnChainPredictionProductV1,
    ProductMetadata as OnChainProductMetadata,
    ProductRegistry,
)
from ..bindings.erc20 import ERC20
from ..bindings.facade import CLEARING_DIAMOND_ABI
from ..bindings.product_registry import ABI as PRODUCT_REGISTRY_ABI
from ..decorators import convert_web3_error
from ..dtos import ExtendedMetadata
from ..exceptions import NotFoundError, ValidationError
from ..schemas import (
    BaseProduct,
    ExpirySpecification,
    OracleSpecification,
    PredictionProduct,
    PredictionProductV1,
    ProductMetadata,
    Transaction,
)
from .base import ClearingSystemAPI, IPFSManager


class Product(ClearingSystemAPI, IPFSManager):
    """API for managing products."""

    def __init__(self, config: Config, authenticator: Authenticator | None = None):
        ClearingSystemAPI.__init__(self, config, authenticator)
        IPFSManager.__init__(self, config)

    ### Product Specification ###

    def validate(self, product_dict: dict[str, Any]) -> PredictionProduct:
        """Creates a product specification from a dictionary.

        The dictionary must follow the schema of the afp.schemas.ProductSpec model.

        Parameters
        ----------
        product_dict : dict
            A dictionary that follows the PredictionProduct schema.

        Returns
        -------
        afp.schemas.PredictionProduct

        Raises
        ------
        ValueError, afp.exceptions.ValidationError
            If the product specification fails validation.
        """
        return self._verify_product_spec(PredictionProduct.model_validate(product_dict))

    def validate_json(self, product_json: str) -> PredictionProduct:
        """Creates a product specification from a dictionary.

        The dictionary must follow the schema of the afp.schemas.ProductSpec model.

        Parameters
        ----------
        product_json : str
            A JSON string that follows the PredictionProduct schema.

        Returns
        -------
        afp.schemas.PredictionProduct

        Raises
        ------
        ValueError, afp.exceptions.ValidationError
            If the product specification fails validation.
        """
        return self._verify_product_spec(
            PredictionProduct.model_validate_json(product_json)
        )

    def dump(self, product_spec: PredictionProduct) -> dict[str, Any]:
        """Creates a dictionary from a product specification.

        Parameters
        ----------
        product_spec : afp.schemas.PredictionProduct
            The product specification.

        Returns
        -------
        dict
        """
        return product_spec.model_dump()

    def dump_json(self, product_spec: PredictionProduct) -> str:
        """Creates a JSON string from a product specification.

        The producted JSON is JSON Canonicalization Scheme (RFC 8785) compliant.

        Parameters
        ----------
        product_spec : afp.schemas.PredictionProduct
            The product specification.

        Returns
        -------
        str
        """
        return product_spec.model_dump_canonical_json()

    def id(self, product_spec: PredictionProduct) -> str:
        """Generates the product ID for a product specification.

        This is the ID that the product will get after successful registration.

        Parameters
        ----------
        product_spec : afp.schemas.PredictionProduct
            The product specification.

        Returns
        -------
        str
        """
        return Web3.to_hex(
            hashing.generate_product_id(
                cast(ChecksumAddress, product_spec.product.base.metadata.builder),
                product_spec.product.base.metadata.symbol,
            )
        )

    ### Transactions ###

    def pin(self, product_spec: PredictionProduct) -> PredictionProduct:
        """Uploads the product's extended metadata to IPFS and pins the CID.

        Adds the CID to the product specification and returns the modified specification
        with the extended metadata CID included.

        Parameters
        ----------
        product_spec : afp.schemas.PredictionProduct or afp.schemas.PredictionProductV1
            The product specification.

        Returns
        -------
        afp.schemas.PredictionProduct
            The product specification with extended metadata CID included.

        Raises
        ------
        afp.exceptions.IPFSError
            If there is an error uploading to the IPFS node.
        """
        extended_metadata_cid = self._ipfs_client.upload_extended_metadata(
            ExtendedMetadata(
                outcome_space=product_spec.outcome_space,
                outcome_point=product_spec.outcome_point,
                oracle_config=product_spec.oracle_config,
                oracle_fallback=product_spec.oracle_fallback,
            )
        )
        updated_base_product = product_spec.product.base.model_copy(
            update=dict(extended_metadata=extended_metadata_cid)
        )
        updated_product = product_spec.product.model_copy(
            update=dict(base=updated_base_product)
        )
        return product_spec.model_copy(update=dict(product=updated_product))

    @convert_web3_error(PRODUCT_REGISTRY_ABI, CLEARING_DIAMOND_ABI)
    def register(
        self, product_spec: PredictionProduct, initial_builder_stake: Decimal
    ) -> Transaction:
        """Submits a product specification to the clearing system.

        The extended metadata should already be pinned to IPFS and the CID should be
        specified in the afp.schemas.BaseProduct object in the product specification.

        Parameters
        ----------
        product_spec : afp.schemas.PredictionProduct or afp.schemas.PredictionProductV1
            The product specification.
        initial_builder_stake : Decimal
            Registration stake (product maintenance fee) in units of the collateral
            asset.

        Returns
        -------
        afp.schemas.Transaction
            Transaction parameters.
        """
        if product_spec.product.base.extended_metadata is None:
            raise ValidationError(
                "Extended metadata CID missing from product specification"
            )

        erc20_contract = ERC20(
            self._w3, cast(ChecksumAddress, product_spec.product.base.collateral_asset)
        )
        decimals = erc20_contract.decimals()

        product_registry_contract = ProductRegistry(
            self._w3, self._config.product_registry_address
        )
        return self._transact(
            product_registry_contract.register_prediction_product(
                self._convert_prediction_product_specification(
                    product_spec.product, decimals
                ),
                int(initial_builder_stake * 10**decimals),
            )
        )

    @convert_web3_error(CLEARING_DIAMOND_ABI)
    def initiate_final_settlement(
        self, product_id: str, accounts: list[str]
    ) -> Transaction:
        """Initiate final settlement (closeout) process for the specified accounts.

        The product must be in Final Settlement state. The accounts must hold non-zero
        positions in the product that offset each other (i.e. the sum of their position
        sizes is 0.)

        Parameters
        ----------
        product_id : str
            The ID of the product.
        accounts : list of str
            List of margin account IDs to initiate settlement for.

        Returns
        -------
        afp.schemas.Transaction
            Transaction parameters.
        """
        product_id = validators.validate_hexstr32(product_id)
        addresses = [validators.validate_address(account) for account in accounts]

        clearing_contract = ClearingDiamond(
            self._w3, self._config.clearing_diamond_address
        )
        return self._transact(
            clearing_contract.initiate_final_settlement(HexBytes(product_id), addresses)
        )

    ### Views ###

    @convert_web3_error(PRODUCT_REGISTRY_ABI, CLEARING_DIAMOND_ABI)
    def get(self, product_id: str) -> PredictionProduct:
        """Retrieves a product registered on chain.

        Parameters
        ----------
        product_id : str
            The ID of the product.

        Returns
        -------
        afp.schemas.PredictionProduct
        """
        product_id = validators.validate_hexstr32(product_id)

        product_registry_contract = ProductRegistry(
            self._w3, self._config.product_registry_address
        )
        # Note: when FuturesProductV1 support is added, call type_of(product_id) first
        # to figure out which view function should be used
        product = product_registry_contract.prediction_product_v1(HexBytes(product_id))

        erc20_contract = ERC20(self._w3, product.base.collateral_asset)
        decimals = erc20_contract.decimals()

        product = self._convert_on_chain_prediction_product(product, decimals)
        if product.base.extended_metadata is None:
            raise ValidationError(
                "Extended metadata CID missing from product specification"
            )

        extended_metadata = self._ipfs_client.download_extended_metadata(
            product.base.extended_metadata
        )
        return PredictionProduct(
            product=product,
            outcome_space=extended_metadata.outcome_space,
            outcome_point=extended_metadata.outcome_point,
            oracle_config=extended_metadata.oracle_config,
            oracle_fallback=extended_metadata.oracle_fallback,
        )

    @convert_web3_error(PRODUCT_REGISTRY_ABI, CLEARING_DIAMOND_ABI)
    def state(self, product_id: str) -> str:
        """Returns the current state of a product.

        Parameters
        ----------
        product_id : str
            The ID of the product.

        Returns
        -------
        str
        """
        product_id = validators.validate_hexstr32(product_id)
        product_registry_contract = ProductRegistry(
            self._w3, self._config.product_registry_address
        )
        state = product_registry_contract.state(HexBytes(product_id))
        return state.name

    @convert_web3_error(PRODUCT_REGISTRY_ABI, CLEARING_DIAMOND_ABI)
    def collateral_asset(self, product_id: str) -> str:
        """Returns the collateral asset of a product.

        Parameters
        ----------
        product_id : str
            The ID of the product.

        Returns
        -------
        str
        """
        product_id = validators.validate_hexstr32(product_id)
        product_registry_contract = ProductRegistry(
            self._w3, self._config.product_registry_address
        )
        collateral_asset = product_registry_contract.collateral_asset(
            HexBytes(product_id)
        )
        if Web3.to_int(hexstr=collateral_asset) == 0:
            raise NotFoundError("Product not found in the product registry")
        return collateral_asset

    ### Internal helpers ###

    def _verify_product_spec(
        self, product_spec: PredictionProduct
    ) -> PredictionProduct:
        validators.verify_collateral_asset(
            self._w3, product_spec.product.base.collateral_asset
        )
        validators.verify_oracle(
            self._w3, product_spec.product.base.oracle_spec.oracle_address
        )
        return product_spec

    @staticmethod
    def _convert_prediction_product_specification(
        product: PredictionProductV1, collateral_asset_decimals: int
    ) -> OnChainPredictionProductV1:
        assert product.base.extended_metadata is not None

        return OnChainPredictionProductV1(
            base=OnChainBaseProduct(
                metadata=OnChainProductMetadata(
                    builder=cast(ChecksumAddress, product.base.metadata.builder),
                    symbol=product.base.metadata.symbol,
                    description=product.base.metadata.description,
                ),
                oracle_spec=OnChainOracleSpecification(
                    oracle_address=cast(
                        ChecksumAddress, product.base.oracle_spec.oracle_address
                    ),
                    fsv_decimals=product.base.oracle_spec.fsv_decimals,
                    fsp_alpha=int(
                        product.base.oracle_spec.fsp_alpha
                        * constants.FULL_PRECISION_MULTIPLIER
                    ),
                    fsp_beta=int(
                        product.base.oracle_spec.fsp_beta
                        * 10**product.base.oracle_spec.fsv_decimals
                    ),
                    fsv_calldata=HexBytes(product.base.oracle_spec.fsv_calldata),
                ),
                collateral_asset=cast(ChecksumAddress, product.base.collateral_asset),
                start_time=int(product.base.start_time.timestamp()),
                point_value=int(
                    product.base.point_value * 10**collateral_asset_decimals
                ),
                price_decimals=product.base.price_decimals,
                extended_metadata=product.base.extended_metadata,
            ),
            expiry_spec=OnChainExpirySpecification(
                earliest_fsp_submission_time=int(
                    product.expiry_spec.earliest_fsp_submission_time.timestamp()
                ),
                tradeout_interval=product.expiry_spec.tradeout_interval,
            ),
            max_price=int(product.max_price * 10**product.base.price_decimals),
            min_price=int(product.min_price * 10**product.base.price_decimals),
        )

    @staticmethod
    def _convert_on_chain_prediction_product(
        product: OnChainPredictionProductV1, collateral_asset_decimals: int
    ) -> PredictionProductV1:
        return PredictionProductV1(
            base=BaseProduct(
                metadata=ProductMetadata(
                    builder=product.base.metadata.builder,
                    symbol=product.base.metadata.symbol,
                    description=product.base.metadata.description,
                ),
                oracle_spec=OracleSpecification(
                    oracle_address=product.base.oracle_spec.oracle_address,
                    fsv_decimals=product.base.oracle_spec.fsv_decimals,
                    fsp_alpha=(
                        Decimal(product.base.oracle_spec.fsp_alpha)
                        / constants.FULL_PRECISION_MULTIPLIER
                    ),
                    fsp_beta=(
                        Decimal(product.base.oracle_spec.fsp_beta)
                        / 10**product.base.oracle_spec.fsv_decimals
                    ),
                    fsv_calldata=product.base.oracle_spec.fsv_calldata.to_0x_hex(),
                ),
                collateral_asset=product.base.collateral_asset,
                start_time=datetime.fromtimestamp(product.base.start_time),
                point_value=(
                    Decimal(product.base.point_value) / 10**collateral_asset_decimals
                ),
                price_decimals=product.base.price_decimals,
                extended_metadata=product.base.extended_metadata,
            ),
            expiry_spec=ExpirySpecification(
                earliest_fsp_submission_time=datetime.fromtimestamp(
                    product.expiry_spec.earliest_fsp_submission_time
                ),
                tradeout_interval=product.expiry_spec.tradeout_interval,
            ),
            max_price=Decimal(product.max_price) / 10**product.base.price_decimals,
            min_price=Decimal(product.min_price) / 10**product.base.price_decimals,
        )
